import numpy as np
import scipy.ndimage


def _steiner_conv_strat(
    refl,
    x,
    y,
    dx,
    dy,
    intense=42,
    peak_relation=0,
    area_relation=1,
    bkg_rad=11000,
    use_intense=True,
):
    """
    We perform the Steiner et al. (1995) algorithm for echo classification
    using only the reflectivity field in order to classify each grid point
    as either convective, stratiform or undefined. Grid points are
    classified as follows,

    0 = Undefined
    1 = Stratiform
    2 = Convective
    """

    def convective_radius(ze_bkg, area_relation):
        """
        Given a mean background reflectivity value, we determine via a step
        function what the corresponding convective radius would be.

        Higher background reflectivitives are expected to have larger
        convective influence on surrounding areas, so a larger convective
        radius would be prescribed.
        """
        if area_relation == 0:
            if ze_bkg < 30:
                conv_rad = 1000.0
            elif (ze_bkg >= 30) & (ze_bkg < 35.0):
                conv_rad = 2000.0
            elif (ze_bkg >= 35.0) & (ze_bkg < 40.0):
                conv_rad = 3000.0
            elif (ze_bkg >= 40.0) & (ze_bkg < 45.0):
                conv_rad = 4000.0
            else:
                conv_rad = 5000.0

        if area_relation == 1:
            if ze_bkg < 25:
                conv_rad = 1000.0
            elif (ze_bkg >= 25) & (ze_bkg < 30.0):
                conv_rad = 2000.0
            elif (ze_bkg >= 30.0) & (ze_bkg < 35.0):
                conv_rad = 3000.0
            elif (ze_bkg >= 35.0) & (ze_bkg < 40.0):
                conv_rad = 4000.0
            else:
                conv_rad = 5000.0

        if area_relation == 2:
            if ze_bkg < 20:
                conv_rad = 1000.0
            elif (ze_bkg >= 20) & (ze_bkg < 25.0):
                conv_rad = 2000.0
            elif (ze_bkg >= 25.0) & (ze_bkg < 30.0):
                conv_rad = 3000.0
            elif (ze_bkg >= 30.0) & (ze_bkg < 35.0):
                conv_rad = 4000.0
            else:
                conv_rad = 5000.0

        if area_relation == 3:
            if ze_bkg < 40:
                conv_rad = 0.0
            elif (ze_bkg >= 40) & (ze_bkg < 45.0):
                conv_rad = 1000.0
            elif (ze_bkg >= 45.0) & (ze_bkg < 50.0):
                conv_rad = 2000.0
            elif (ze_bkg >= 50.0) & (ze_bkg < 55.0):
                conv_rad = 6000.0
            else:
                conv_rad = 8000.0

        return conv_rad

    def peakedness(ze_bkg, peak_relation):
        """
        Given a background reflectivity value, we determine what the necessary
        peakedness (or difference) has to be between a grid point's
        reflectivity and the background reflectivity in order for that grid
        point to be labeled convective.
        """
        if peak_relation == 0:
            if ze_bkg < 0.0:
                peak = 10.0
            elif (ze_bkg >= 0.0) and (ze_bkg < 42.43):
                peak = 10.0 - ze_bkg**2 / 180.0
            else:
                peak = 0.0

        elif peak_relation == 1:
            if ze_bkg < 0.0:
                peak = 14.0
            elif (ze_bkg >= 0.0) and (ze_bkg < 42.43):
                peak = 14.0 - ze_bkg**2 / 180.0
            else:
                peak = 4.0

        return peak

    sclass = np.zeros(refl.shape, dtype=int)
    ny, nx = refl.shape

    for i in range(0, nx):
        # Get stencil of x grid points within the background radius
        imin = np.max(np.array([1, (i - bkg_rad / dx)], dtype=int))
        imax = np.min(np.array([nx, (i + bkg_rad / dx)], dtype=int))

        for j in range(0, ny):
            # First make sure that the current grid point has not already been
            # classified. This can happen when grid points within the
            # convective radius of a previous grid point have also been
            # classified.
            if ~np.isnan(refl[j, i]) & (sclass[j, i] == 0):
                # Get stencil of y grid points within the background radius
                jmin = np.max(np.array([1, (j - bkg_rad / dy)], dtype=int))
                jmax = np.min(np.array([ny, (j + bkg_rad / dy)], dtype=int))

                n = 0
                sum_ze = 0

                # Calculate the mean background reflectivity for the current
                # grid point, which will be used to determine the convective
                # radius and the required peakedness.

                for r in range(imin, imax):
                    for m in range(jmin, jmax):
                        if not np.isnan(refl[m, r]):
                            rad = np.sqrt((x[r] - x[i]) ** 2 + (y[m] - y[j]) ** 2)

                            # The mean background reflectivity will first be
                            # computed in linear units, i.e. mm^6/m^3, then
                            # converted to decibel units.
                            if rad <= bkg_rad:
                                n += 1
                                sum_ze += 10.0 ** (refl[m, r] / 10.0)

                if n == 0:
                    ze_bkg = np.inf
                else:
                    ze_bkg = 10.0 * np.log10(sum_ze / n)

                # Now get the corresponding convective radius knowing the mean
                # background reflectivity.
                conv_rad = convective_radius(ze_bkg, area_relation)

                # Now we want to investigate the points surrounding the current
                # grid point that are within the convective radius, and whether
                # they too are convective, stratiform or undefined.

                # Get stencil of x and y grid points within the convective
                # radius.
                lmin = np.max(np.array([1, int(i - conv_rad / dx)], dtype=int))
                lmax = np.min(np.array([nx, int(i + conv_rad / dx)], dtype=int))
                mmin = np.max(np.array([1, int(j - conv_rad / dy)], dtype=int))
                mmax = np.min(np.array([ny, int(j + conv_rad / dy)], dtype=int))

                if use_intense and (refl[j, i] >= intense):
                    sclass[j, i] = 2

                    for r in range(lmin, lmax):
                        for m in range(mmin, mmax):
                            if not np.isnan(refl[m, r]):
                                rad = np.sqrt((x[r] - x[i]) ** 2 + (y[m] - y[j]) ** 2)

                                if rad <= conv_rad:
                                    sclass[m, r] = 2

                else:
                    peak = peakedness(ze_bkg, peak_relation)

                    if refl[j, i] - ze_bkg >= peak:
                        sclass[j, i] = 2

                        for r in range(imin, imax):
                            for m in range(jmin, jmax):
                                if not np.isnan(refl[m, r]):
                                    rad = np.sqrt(
                                        (x[r] - x[i]) ** 2 + (y[m] - y[j]) ** 2
                                    )

                                    if rad <= conv_rad:
                                        sclass[m, r] = 2

                    else:
                        # If by now the current grid point has not been
                        # classified as convective by either the intensity
                        # criteria or the peakedness criteria, then it must be
                        # stratiform.
                        sclass[j, i] = 1

    return sclass


def steiner_class_buff(
    ze,
    x,
    y,
    z,
    dx,
    dy,
    bkg_rad,
    work_level,
    intense,
    peak_relation,
    area_relation,
    use_intense,
):

    zslice = np.argmin(np.abs(z - work_level))
    refl = ze[zslice, :, :]

    area_rel = {"small": 0, "medium": 1, "large": 2, "sgp": 3}
    peak_rel = {"default": 0, "sgp": 1}

    sclass = _steiner_conv_strat(
        refl,
        x,
        y,
        dx,
        dy,
        intense=intense,
        peak_relation=peak_rel[peak_relation],
        area_relation=area_rel[area_relation],
        bkg_rad=11000,
        use_intense=True,
    )

    return sclass


def _revised_conv_strat(
    refl,
    dx,
    dy,
    always_core_thres=42,
    bkg_rad_km=11,
    use_cosine=True,
    max_diff=5,
    zero_diff_cos_val=55,
    scalar_diff=1.5,
    use_addition=True,
    calc_thres=0.75,
    weak_echo_thres=5.0,
    min_dBZ_used=5.0,
    dB_averaging=True,
    remove_small_objects=True,
    min_km2_size=10,
    val_for_max_conv_rad=30,
    max_conv_rad_km=5.0,
    cs_core=3,
    nosfcecho=0,
    weakecho=3,
    sf=1,
    conv=2,
):
    """
    We perform the Yuter and Houze (1997) algorithm for echo classification
    using only the reflectivity field in order to classify each grid point
    as either convective, stratiform or undefined. Grid points are
    classified as follows,

    nosfcecho = No Surface Echo/ Undefined
    sf = Stratiform
    conv = Convective
    weakecho = Weak Echo

    refl : array
        array of reflectivity values
    x, y : array
        x and y coordinates of reflectivity array, respectively
    dx, dy : float
        The x- and y-dimension resolutions in meters, respectively.
    always_core_thres : float, optional
        Threshold for points that are always convective. All values above the threshold are classifed as convective
    bkg_rad_km : float, optional
        Radius to compute background reflectivity in kilometers. Default is 11 km. Recommended to be at least 3 x
        grid spacing
    use_cosine : bool, optional
        Boolean used to determine if cosine scheme should be used for identifying convective cores (True) or a scalar
        scheme (False)
    max_diff : float, optional
        Maximum difference between background average and reflectivity in order to be classified as convective.
        "a" value in Eqn. B1 in Yuter and Houze (1997)
    zero_diff_cos_val : float, optional
        Value where difference between background average and reflectivity is zero in the cosine function
        "b" value in Eqn. B1 in Yuter and Houze (1997)
    scalar_diff : float, optional
        If using a scalar difference scheme, this value is the multiplier or addition to the background average
    use_addition : bool, optional
        Determines if a multiplier (False) or addition (True) in the scalar difference scheme should be used
    calc_thres : float, optional
        Minimum percentage of points needed to be considered in background average calculation
    weak_echo_thres : float, optional
        Threshold for determining weak echo. All values below this threshold will be considered weak echo
    min_dBZ_used : float, optional
        Minimum dBZ value used for classification. All values below this threshold will be considered no surface echo
    dB_averaging : bool, optional
        True if using dBZ values that need to be converted to linear Z before averaging. False for other types of values
    remove_small_objects : bool, optional
        Determines if small objects should be removed from convective core array. Default is True.
    min_km2_size : float, optional
        Minimum size of convective cores to be considered. Cores less than this size will be removed. Default is 10
        km^2.
    val_for_max_conv_rad : float, optional
        dBZ for maximum convective radius. Convective cores with values above this will have the maximum convective
        radius
    max_conv_rad_km : float, optional
        Maximum radius around convective cores to classify as convective. Default is 5 km.
    cs_core : int, optional
        Value for points classified as convective cores
    nosfcecho : int, optional
        Value for points classified as no surface echo, based on min_dBZ_used
    weakecho : int, optional
        Value for points classified as weak echo, based on weak_echo_thres
    sf : int, optional
        Value for points classified as stratiform
    conv : int, optional
        Value for points classified as convective

    Returns
    -------
    refl_bkg : array
        Array of background values
    conv_core_array : array
        Array of initial convective cores (identified convective elements without convective radii applied)
    conv_strat_array : array
        Array of convective stratiform classifcation with convective radii applied
    """

    # Set up mask arrays for background average and
    # prepare for convective mask arrays
    # calculate maximum convective diameter from max. convective radius (input)
    max_conv_diameter = int(np.floor((max_conv_rad_km / (dx / 1000)) * 2))
    # if diameter is even, make odd
    if max_conv_diameter % 2 == 0:
        max_conv_diameter = max_conv_diameter + 1
    # find center point
    center_conv_mask_x = int(np.floor(max_conv_diameter / 2))

    # prepare background mask array for computing background average
    # calculate number of pixels for background array given requested background radius and dx
    bkg_diameter_pix = int(np.floor((bkg_rad_km / (dx / 1000)) * 2))
    # set diameter to odd if even
    if bkg_diameter_pix % 2 == 0:
        bkg_diameter_pix = bkg_diameter_pix + 1
    # find center point
    bkg_center = int(np.floor(bkg_diameter_pix / 2))
    # create background array
    bkg_mask_array = np.ones((bkg_diameter_pix, bkg_diameter_pix), dtype=float)
    # mask outside circular region
    bkg_mask_array = create_radial_mask(
        bkg_mask_array,
        min_rad_km=0,
        max_rad_km=bkg_rad_km,
        x_pixsize=dx / 1000,
        y_pixsize=dy / 1000,
        center_x=bkg_center,
        center_y=bkg_center,
        circular=True,
    )

    # Convective stratiform detection
    # start by making reflectivity a masked array
    refl = np.ma.masked_invalid(refl)
    # Compute background radius
    refl_bkg = calc_bkg_intensity(refl, bkg_mask_array, dB_averaging, calc_thres)
    # mask reflectivity field
    refl = np.ma.masked_where(refl_bkg.mask, refl)

    # Get convective core array from cosine scheme, or scalar scheme
    if use_cosine:
        conv_core_array = convcore_cos_scheme(
            refl, refl_bkg, max_diff, zero_diff_cos_val, always_core_thres, cs_core
        )
    else:
        conv_core_array = convcore_scalar_scheme(
            refl,
            refl_bkg,
            scalar_diff,
            always_core_thres,
            cs_core,
            use_addition=use_addition,
        )

    # Assign convective radii based on background reflectivity
    conv_radius_km = assign_conv_radius_km(
        refl_bkg,
        val_for_max_conv_rad=val_for_max_conv_rad,
        max_conv_rad=max_conv_rad_km,
    )

    # remove small objects in convective core array
    if remove_small_objects:
        # calculate minimum pixel size given dx and dy
        min_pix_size = min_km2_size / ((dx / 1000) * (dy / 1000))
        # label connected objects in convective core array
        cc_labels, _ = scipy.ndimage.label(conv_core_array)
        # mask labels where convective core array is masked
        cc_labels = np.ma.masked_where(conv_core_array.mask, cc_labels)

        # loop through each unique label
        for lab in np.unique(cc_labels):
            # calculate number of pixels for each label
            size_lab = np.count_nonzero(cc_labels == lab)
            # if number of pixels is less than minimum, then remove core
            if size_lab < min_pix_size:
                conv_core_array[cc_labels == lab] = 0

    # Incorporate convective radius using binary dilation
    # Create empty array for assignment
    temp_assignment = np.zeros_like(conv_core_array)

    # Loop through radii
    for radius in np.arange(1, max_conv_rad_km + 1):
        # create mask array for radius incorporation
        conv_mask_array = create_conv_radius_mask(
            max_conv_diameter, radius, dx / 1000, dy / 1000, center_conv_mask_x
        )
        # find location of radius
        temp = conv_radius_km == radius
        # get cores for given radius
        temp_core = np.ma.masked_where(~temp, conv_core_array)
        # dilate cores
        temp_dilated = scipy.ndimage.binary_dilation(
            temp_core.filled(0), conv_mask_array
        )
        # add to assignment array
        temp_assignment = temp_assignment + temp_dilated

    # add dilated cores to original array
    conv_core_copy = np.ma.copy(conv_core_array)
    conv_core_copy[temp_assignment >= 1] = cs_core

    # Now do convective stratiform classification
    conv_strat_array = np.zeros_like(refl)
    conv_strat_array = classify_conv_strat_array(
        refl,
        conv_strat_array,
        conv_core_copy,
        nosfcecho,
        conv,
        sf,
        weakecho,
        cs_core,
        min_dBZ_used,
        weak_echo_thres,
    )
    # mask where reflectivity is masked
    conv_strat_array = np.ma.masked_where(refl.mask, conv_strat_array)

    return refl_bkg, conv_core_array, conv_strat_array


# functions


def create_radial_mask(
    mask_array,
    min_rad_km,
    max_rad_km,
    x_pixsize,
    y_pixsize,
    center_x,
    center_y,
    circular=True,
):
    """
    Computes a radial distance mask, everything with distance between minradiuskm
    and maxradiuskm is assigned 1, everything else is assigned 0. This version can
    handle rectangular arrays and pixels as well as square ones.

    Parameters
    ----------
    mask_array : array
        Array to mask
    min_rad_km, max_rad_km : float
        The minimum and maximum radius of the non-masked region in kilometers.
    x_pixsize, y_pixsize : float
        The pixel size in the x- and y-dimension in kilometers, respectively
    center_x, center_y : int
        The center pixel in the x- and y-dimension, respectively
    circular : bool
        True returns circular mask, False returns a rectangular mask.

    Returns
    -------
    mask_array : array
        Rectangular array masked by a radial distance.
    """

    xsize, ysize = mask_array.shape

    for j in np.arange(0, ysize, 1):
        for i in np.arange(0, xsize, 1):
            # compute range to pixel
            if circular:
                x_range_sq = ((center_x - i) * x_pixsize) ** 2
                y_range_sq = ((center_y - j) * y_pixsize) ** 2
                circ_range = np.sqrt(x_range_sq + y_range_sq)
            # if circular is False, use square mask
            else:
                x_range = abs(int(np.floor(center_x - i) * x_pixsize))
                y_range = abs(int(np.floor(center_y - j) * y_pixsize))

                if x_range > y_range:
                    circ_range = x_range
                else:
                    circ_range = y_range
            # if range is within min and max, set to True
            if (circ_range <= max_rad_km) and (circ_range >= min_rad_km):
                mask_array[j, i] = 1
            else:
                mask_array[j, i] = 0

    return mask_array


def calc_bkg_intensity(refl, bkg_mask_array, dB_averaging, calc_thres=None):
    """
    Calculate the background of the given refl array. The footprint used to
    calculate the average for each pixel is given by bkg_mask_array

    Parameters
    ----------
    refl : array
        Reflectivity array to compute average
    bkg_mask_array : array
        Array of radial points to use for average
    dB_averaging : bool
        If True, converts dBZ to linear Z before averaging
    calc_thres : float
        Minimum percentage of points needed to be considered in background average calculation

    Returns
    -------
    refl_bkg : array
        Array of average values
    """

    # if dBaverage is true, convert reflectivity to linear Z
    if dB_averaging:
        refl = 10 ** (refl / 10)

    # calculate background reflectivity with circular footprint
    refl_bkg = scipy.ndimage.generic_filter(
        refl.filled(np.nan),
        function=np.nanmean,
        mode="constant",
        footprint=bkg_mask_array.astype(bool),
        cval=np.nan,
    )

    # if calc_thres is not none, then calculate the number of points used to calculate average
    if calc_thres is not None:
        # count valid points
        refl_count = scipy.ndimage.generic_filter(
            refl.filled(0),
            function=np.count_nonzero,
            mode="constant",
            footprint=bkg_mask_array.astype(bool),
            cval=0,
        )
        # find threshold number of points
        val = calc_thres * np.count_nonzero(bkg_mask_array)
        # mask out values
        refl_bkg = np.ma.masked_where(refl_count < val, refl_bkg)

    # mask where original reflectivity is invalid
    refl_bkg = np.ma.masked_where(refl.mask, refl_bkg)

    # if dBaveraging is true, convert background reflectivity to dBZ
    if dB_averaging:
        refl_bkg = 10 * np.log10(refl_bkg)
        # mask where original reflectivity is invalid
        refl_bkg = np.ma.masked_where(refl.mask, refl_bkg)

    return refl_bkg


def convcore_cos_scheme(
    refl, refl_bkg, max_diff, zero_diff_cos_val, always_core_thres, CS_CORE
):
    """
    Function for assigning convective cores based on a cosine function

    Parameters
    ----------
    refl : array
        Reflectivity values
    refl_bkg : array
        Background average of reflectivity values
    max_diff : float
        Maximum difference between refl and refl_bkg needed for convective classification
    zero_diff_cos_val : float
        Value where the cosine function returns a zero difference
    always_core_thres : float
        All values above this threshold considered to be convective
    CS_CORE : int
        Value assigned to convective pixels

    Returns
    -------
    conv_core_array : array
        Array of booleans if point is convective (1) or not (0)
    """

    # initialize entire array to not a convective core
    conv_core_array = np.zeros_like(refl)

    # calculate zeDiff for entire array
    zDiff = max_diff * np.cos((np.pi * refl_bkg) / (2 * zero_diff_cos_val))
    zDiff[zDiff < 0] = 0  # where difference less than zero, set to zero
    zDiff[refl_bkg < 0] = max_diff  # where background less than zero, set to max. diff

    # set values
    conv_core_array[
        refl >= always_core_thres
    ] = CS_CORE  # where Z is greater than always_core_thres, set to core
    conv_core_array[
        (refl - refl_bkg) >= zDiff
    ] = CS_CORE  # where difference exceeeds minimum, set to core

    return conv_core_array


def convcore_scalar_scheme(
    refl, refl_bkg, max_diff, always_core_thres, CS_CORE, use_addition=False
):
    """
    Function for assigning convective cores based on a scalar difference

    Parameters
    ----------
    refl : array
        Reflectivity values
    refl_bkg : array
        Background average of reflectivity values
    max_diff : float
        Maximum difference between refl and refl_bkg needed for convective classification
    always_core_thres : float
        All values above this threshold considered to be convective
    CS_CORE : int
        Value assigned to convective pixels
    use_addition : bool
        Boolean to determine if scalar should be added (True) or multiplied (False)

    Returns
    -------
    conv_core_array : array
        Array of booleans if point is convective (1) or not (0)
    """

    # initialize entire array to not a convective core
    conv_core_array = np.zeros_like(refl)

    # calculate zDiff for entire array
    # if addition, add difference. Else, multiply difference
    if use_addition:
        zDiff = max_diff + refl_bkg
    else:
        zDiff = max_diff * refl_bkg

    zDiff[zDiff < 0] = 0  # where difference less than zero, set to zero
    zDiff[refl_bkg < 0] = 0  # where background less than zero, set to zero

    # set values
    conv_core_array[
        refl >= always_core_thres
    ] = CS_CORE  # where Z is greater than always_core_thres, set to core
    conv_core_array[
        refl >= zDiff
    ] = CS_CORE  # where difference exceeeds minimum, set to core

    return conv_core_array


def create_conv_radius_mask(
    max_conv_diameter, radius_km, x_spacing, y_spacing, center_conv_mask_x
):
    """
    Does and initial convective stratiform classification

    Parameters
    ----------
    max_conv_diameter : int
        maximum convective diameter in kilometers
    radius_km : int
        convective radius in kilometers
    x_spacing, y_spacing : float
        x- and y-dimension pixel size in meters, respectively
    center_conv_mask_x : int
        index of center point

    Returns
    -------
    conv_mask_array : array
        array masked based on distance of convective diameter
    """

    conv_mask_array = np.zeros((max_conv_diameter, max_conv_diameter))
    conv_mask_array = create_radial_mask(
        conv_mask_array,
        0,
        radius_km,
        x_spacing,
        y_spacing,
        center_conv_mask_x,
        center_conv_mask_x,
        True,
    )

    return conv_mask_array


def assign_conv_radius_km(refl_bkg, val_for_max_conv_rad, max_conv_rad=5):
    """
    Assigns the convective radius in kilometers based on the background values

    Parameters
    ----------
    refl_bkg : array
        array of background reflectivity values
    val_for_max_conv_rad : float
        reflectivity value for maximum convective radius (5 km)
    max_conv_rad : float, optional
        maximum convective radius in kilometers

    Returns
    -------
    convRadiuskm : array
        array of convective radii based on background values and val for max. conv radius
    """

    convRadiuskm = np.ones_like(refl_bkg)

    convRadiuskm[refl_bkg >= (val_for_max_conv_rad - 15)] = max_conv_rad - 3
    convRadiuskm[refl_bkg >= (val_for_max_conv_rad - 10)] = max_conv_rad - 2
    convRadiuskm[refl_bkg >= (val_for_max_conv_rad - 5)] = max_conv_rad - 1
    convRadiuskm[refl_bkg >= val_for_max_conv_rad] = max_conv_rad

    return convRadiuskm


def classify_conv_strat_array(
    refl,
    conv_strat_array,
    conv_core_array,
    NOSFCECHO,
    CONV,
    SF,
    WEAKECHO,
    CS_CORE,
    MINDBZUSE,
    WEAKECHOTHRES,
):
    """
    Does and initial convective stratiform classification

    Parameters
    ----------
    refl : array
        Array of reflectivity values
    conv_strat_array : array
        Array with convective stratiform classifications
    conv_core_array : array
        Array with convective cores
    NOSFCECHO : int
        Value to assign points classified as no surface echo
    CONV : int
        Value to assign points classified as convective
    SF : int
        Value to assign points classified as stratiform
    WEAKECHO : int
        Value to assign points classfied as weak echo
    CS_CORE : int
        Value assigned to convective cores in conv_core_array
    MINDBZUSE : float
        Minimum dBZ value to consider in classification, all values below this will be set to NOSFCECHO
    WEAKECHOTHRES : float
        dBZ threshold for weak echo classification, all values below this will be set to WEAKECHO

    Returns
    -------
    conv_strat_array : array
        array of classifications
    """

    # assuming order so that each point is only assigned one time, no overlapping assignment
    # initially, assign every point to stratiform
    conv_strat_array[:] = SF
    # where reflectivity is masked, set to no surface echo
    conv_strat_array[refl.mask] = NOSFCECHO
    # assign convective cores to CONV
    conv_strat_array[conv_core_array == CS_CORE] = CONV
    # assign reflectivity less than weakechothres to weak echo
    conv_strat_array[refl < WEAKECHOTHRES] = WEAKECHO
    # assign reflectivity less than minimum to no surface echo
    conv_strat_array[refl < MINDBZUSE] = NOSFCECHO

    return conv_strat_array



"""
Py-ART has a robust function for mapping radar data from the collected radar
coordinates to Cartesian coordinates.

"""

from .gate_mapper import GateMapper  # noqa
from .gates_to_grid import map_gates_to_grid  # noqa
from .grid_mapper import example_roi_func_constant  # noqa
from .grid_mapper import example_roi_func_dist  # noqa
from .grid_mapper import example_roi_func_dist_beam  # noqa
from .grid_mapper import grid_from_radars  # noqa
from .grid_mapper import map_to_grid  # noqa

__all__ = [s for s in dir() if not s.startswith("_")]



""" Unit Tests for Py-ART's io/grid_io.py module. """

import netCDF4
import numpy as np
from numpy.testing import assert_almost_equal, assert_warns

import pyart
from pyart.io.common import stringarray_to_chararray


def test_grid_write_read():
    # test the read_grid and write_grid function by performing a
    # write/read roundtrip and comparing the two Grid objects
    grid1 = pyart.testing.make_target_grid()
    grid1.projection["comment"] = "This is a comment"
    grid1.metadata["comment"] = "This is another comment"

    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_grid.nc"
        pyart.io.write_grid(tmpfile, grid1)
        grid2 = pyart.io.read_grid(tmpfile)

        # check fields
        for field in grid1.fields.keys():
            _check_dicts_similar(grid1.fields[field], grid2.fields[field])

        # check attributes
        assert "Conventions" in grid2.metadata
        grid2.metadata.pop("Conventions")
        _check_attrs_similar(grid1, grid2, "metadata")

        _check_attrs_similar(grid1, grid2, "time")

        _check_attrs_similar(grid1, grid2, "origin_latitude")
        _check_attrs_similar(grid1, grid2, "origin_longitude")
        _check_attrs_similar(grid1, grid2, "origin_altitude")

        _check_attrs_similar(grid1, grid2, "x")
        _check_attrs_similar(grid1, grid2, "y")
        _check_attrs_similar(grid1, grid2, "z")

        _check_attrs_similar(grid1, grid2, "point_x")
        _check_attrs_similar(grid1, grid2, "point_y")
        _check_attrs_similar(grid1, grid2, "point_z")

        _check_attrs_similar(grid1, grid2, "projection")
        assert grid1.projection["_include_lon_0_lat_0"] is True

        _check_attrs_similar(grid1, grid2, "point_latitude")
        _check_attrs_similar(grid1, grid2, "point_longitude")
        _check_attrs_similar(grid1, grid2, "point_altitude")

        assert grid1.nx == grid2.nx
        assert grid1.ny == grid2.ny
        assert grid1.nz == grid2.nz

        _check_attrs_similar(grid1, grid2, "radar_latitude")
        _check_attrs_similar(grid1, grid2, "radar_longitude")
        _check_attrs_similar(grid1, grid2, "radar_altitude")
        _check_attrs_similar(grid1, grid2, "radar_time")
        _check_attrs_similar(grid1, grid2, "radar_name")
        assert grid1.nradar == grid2.nradar


def _check_attrs_similar(grid1, grid2, attr):
    print("Checking attribute:", attr)
    dic1 = getattr(grid1, attr)
    dic2 = getattr(grid2, attr)
    _check_dicts_similar(dic1, dic2)


def _check_dicts_similar(dic1, dic2):
    for k, v in dic1.items():
        print("Checking key:", k)
        if k == "data":
            if v.dtype.char == "S" or v.dtype.char == "U":
                s1 = v.astype("S")
                if s1.ndim == 1:
                    s1 = stringarray_to_chararray(s1)

                s2 = dic2[k].astype("S")
                if s1.ndim == 1:
                    s2 = stringarray_to_chararray(s2)

                assert b"".join(*s1) == b"".join(*s2)
            else:
                assert_almost_equal(v, dic2[k])
        else:
            assert dic2[k] == v


def test_write_grid_fields_list():
    grid1 = pyart.testing.make_target_grid()
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_grid.nc"
        tmpfile_warn = "tmp_ppi_warn.nc"
        include_fields = ["reflectivity"]
        _format = "NETCDF4"
        pyart.io.write_grid(tmpfile, grid1, include_fields=include_fields)
        grid2 = pyart.io.read_grid(tmpfile)
        assert "reflectivity" in grid2.fields.keys()
        include_fields = ["foo"]
        assert_warns(
            UserWarning,
            pyart.io.write_grid,
            tmpfile_warn,
            grid1,
            _format,
            include_fields,
        )


def test_grid_write_point_vars():
    grid1 = pyart.testing.make_target_grid()

    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_grid.nc"
        pyart.io.write_grid(
            tmpfile, grid1, write_point_x_y_z=True, write_point_lon_lat_alt=True
        )
        dset = netCDF4.Dataset(tmpfile, "r")

        assert "point_x" in dset.variables
        assert "point_y" in dset.variables
        assert "point_z" in dset.variables
        assert "point_latitude" in dset.variables
        assert "point_longitude" in dset.variables
        assert "point_altitude" in dset.variables
        dset.close()


def test_grid_write_arm_time_vars():
    grid1 = pyart.testing.make_target_grid()
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_grid.nc"
        pyart.io.write_grid(tmpfile, grid1, arm_time_variables=True)
        dset = netCDF4.Dataset(tmpfile, "r")
        assert "base_time" in dset.variables
        assert "time_offset" in dset.variables
        dset.close()


def test_exclude_fields_argument():
    grid1 = pyart.testing.make_target_grid()
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_grid.nc"
        pyart.io.write_grid(tmpfile, grid1)
        grid2 = pyart.io.read_grid(tmpfile, exclude_fields=["reflectivity"])
        assert "reflectivity" in grid1.fields
        assert "reflectivity" not in grid2.fields


def test_radar_attrs_none():
    grid1 = pyart.testing.make_target_grid()
    grid1.radar_latitude = None
    grid1.radar_longitude = None
    grid1.radar_altitude = None
    grid1.radar_time = None
    grid1.radar_name = None
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_grid.nc"
        pyart.io.write_grid(tmpfile, grid1)
        grid2 = pyart.io.read_grid(tmpfile)
        assert grid2.radar_latitude is None
        assert grid2.radar_longitude is None
        assert grid2.radar_altitude is None
        assert grid2.radar_time is None
        assert grid2.radar_name is None


def test_bad_shaped_field():
    grid1 = pyart.testing.make_target_grid()
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_grid.nc"
        pyart.io.write_grid(tmpfile, grid1)

        # add a scalar variable named bad_field
        dset = netCDF4.Dataset(tmpfile, "a")
        dset.createVariable("bad_field", "f4", ())
        dset.close()

        # warning should be raised about incorrect shape of bad_field var
        assert_warns(UserWarning, pyart.io.read_grid, tmpfile)


def test_write_projection_coordinate_system():
    grid_shape = (2, 3, 4)
    grid_limits = ((0, 500), (-400000, 400000), (-300000, 300000))
    grid = pyart.testing.make_empty_grid(grid_shape, grid_limits)

    with pyart.testing.InTemporaryDirectory():

        # standard projection
        grid.projection["proj"] = "pyart_aeqd"
        tmpfile = "tmp_grid1.nc"
        pyart.io.write_grid(tmpfile, grid, write_proj_coord_sys=True)
        dset = netCDF4.Dataset(tmpfile, "r")
        assert "ProjectionCoordinateSystem" in dset.variables
        dset.close()

        # explicit proj_coord_sys
        proj_coord_sys = {"foo": "bar"}
        tmpfile = "tmp_grid2.nc"
        pyart.io.write_grid(
            tmpfile, grid, write_proj_coord_sys=True, proj_coord_sys=proj_coord_sys
        )
        dset = netCDF4.Dataset(tmpfile, "r")
        assert "ProjectionCoordinateSystem" in dset.variables
        assert dset.variables["ProjectionCoordinateSystem"].foo == "bar"
        dset.close()

        # explicit proj_coord_sys
        proj_coord_sys = {"foo": "bar"}
        tmpfile = "tmp_grid3.nc"
        pyart.io.write_grid(tmpfile, grid, write_proj_coord_sys=False)
        dset = netCDF4.Dataset(tmpfile, "r")
        assert "ProjectionCoordinateSystem" not in dset.variables
        dset.close()


def test_unknown_projection():
    grid_shape = (2, 3, 4)
    grid_limits = ((0, 500), (-400000, 400000), (-300000, 300000))
    grid = pyart.testing.make_empty_grid(grid_shape, grid_limits)
    grid.projection["proj"] = "null"

    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_grid.nc"
        assert_warns(
            UserWarning, pyart.io.write_grid, tmpfile, grid, write_proj_coord_sys=True
        )
        dset = netCDF4.Dataset(tmpfile, "r")
        assert "ProjectionCoordinateSystem" not in dset.variables
        dset.close()


def test_make_coordinate_system_dict():
    grid_shape = (2, 3, 4)
    grid_limits = ((0, 500), (-400000, 400000), (-300000, 300000))
    grid = pyart.testing.make_empty_grid(grid_shape, grid_limits)

    # pyart_aeqd
    grid.projection["proj"] = "pyart_aeqd"
    dic = pyart.io.grid_io._make_coordinatesystem_dict(grid)
    assert dic["grid_mapping_name"] == "azimuthal_equidistant"
    assert "semi_major_axis" in dic
    assert "inverse_flattening" in dic
    assert "longitude_of_prime_meridian" in dic
    assert "false_easting" in dic
    assert "false_northing" in dic
    assert "longitude_of_projection_origin" in dic
    assert "latitude_of_projection_origin" in dic

    # aeqd
    grid.projection["proj"] = "aeqd"
    dic = pyart.io.grid_io._make_coordinatesystem_dict(grid)
    assert dic["grid_mapping_name"] == "azimuthal_equidistant"
    assert "semi_major_axis" in dic
    assert "inverse_flattening" in dic
    assert "longitude_of_prime_meridian" in dic
    assert "false_easting" in dic
    assert "false_northing" in dic
    assert "longitude_of_projection_origin" in dic
    assert "latitude_of_projection_origin" in dic

    # tmerc
    grid.projection["proj"] = "tmerc"
    dic = pyart.io.grid_io._make_coordinatesystem_dict(grid)
    assert dic["grid_mapping_name"] == "transverse_mercator"
    assert "longitude_of_central_meridian" in dic
    assert "latitude_of_projection_origin" in dic
    assert "scale_factor_at_central_meridian" in dic

    # lcc
    grid.projection["proj"] = "lcc"
    dic = pyart.io.grid_io._make_coordinatesystem_dict(grid)
    assert dic["grid_mapping_name"] == "lambert_conformal_conic"
    assert "standard_parallel" in dic
    assert "longitude_of_projection_origin" in dic
    assert "latitude_of_projection_origin" in dic

    # laea
    grid.projection["proj"] = "laea"
    dic = pyart.io.grid_io._make_coordinatesystem_dict(grid)
    assert dic["grid_mapping_name"] == "lambert_azimuthal_equal_area"
    assert "longitude_of_projection_origin" in dic
    assert "latitude_of_projection_origin" in dic

    # aea
    grid.projection["proj"] = "aea"
    dic = pyart.io.grid_io._make_coordinatesystem_dict(grid)
    assert dic["grid_mapping_name"] == "albers_conical_equal_area"
    assert "standard_parallel" in dic
    assert "longitude_of_central_meridian" in dic
    assert "latitude_of_projection_origin" in dic

    # stere
    grid.projection["proj"] = "stere"
    dic = pyart.io.grid_io._make_coordinatesystem_dict(grid)
    assert dic["grid_mapping_name"] == "stereographic"
    assert "longitude_of_projection_origin" in dic
    assert "latitude_of_projection_origin" in dic
    assert "scale_factor_at_projection_origin" in dic

    # npstere
    grid.projection["proj"] = "npstere"
    dic = pyart.io.grid_io._make_coordinatesystem_dict(grid)
    assert dic["grid_mapping_name"] == "polar_stereographic"
    assert "longitude_of_projection_origin" in dic
    assert "latitude_of_projection_origin" in dic
    assert "standard_parallel" in dic

    # spstere
    grid.projection["proj"] = "npstere"
    dic = pyart.io.grid_io._make_coordinatesystem_dict(grid)
    assert dic["grid_mapping_name"] == "polar_stereographic"
    assert "longitude_of_projection_origin" in dic
    assert "latitude_of_projection_origin" in dic
    assert "standard_parallel" in dic

    # ortho
    grid.projection["proj"] = "ortho"
    dic = pyart.io.grid_io._make_coordinatesystem_dict(grid)
    assert dic["grid_mapping_name"] == "orthographic"
    assert "longitude_of_projection_origin" in dic
    assert "latitude_of_projection_origin" in dic

    # projection that does not map to a CDM transform
    grid.projection["proj"] = "null"
    dic = pyart.io.grid_io._make_coordinatesystem_dict(grid)
    assert dic is None


def test_write_grid_empty_radar_names():
    # GitHub issue #537
    grid1 = pyart.testing.make_target_grid()
    grid1.nradar = 2
    grid1.radar_name["data"] = np.array(["", ""])
    grid1.radar_latitude = None
    grid1.radar_longitude = None
    grid1.radar_altitude = None
    grid1.radar_time = None

    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_grid.nc"
        pyart.io.write_grid(tmpfile, grid1)



"""
=========================================
Calculate and Plot Composite Reflectivity
=========================================

Calculates and plots the composite reflectivity, or the
maximum reflectivity across all of the elevations.
"""

# Author: Maxwell Grover (mgrover@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt

import pyart
from pyart.testing import get_test_data

# Read in a sample file
filename = get_test_data("swx_20120520_0641.nc")
radar = pyart.io.read(filename)

# Configure a gatefilter to filter out copolar correlation coefficient values > 0.9
gatefilter = pyart.filters.GateFilter(radar)
gatefilter.exclude_transition()
gatefilter.exclude_below("copol_coeff", 0.9)

# Calculate composite reflectivity, or the maximum reflectivity across all elevation levels
compz = pyart.retrieve.composite_reflectivity(
    radar, field="reflectivity_horizontal", gatefilter=gatefilter
)

# Plot the original reflectivity field and the composite field
fig = plt.figure(figsize=(16, 6))
ax = plt.subplot(121)
display = pyart.graph.RadarDisplay(radar)
display.plot("reflectivity_horizontal", ax=ax, vmin=-20, vmax=80)

ax2 = plt.subplot(122)
composite_display = pyart.graph.RadarDisplay(compz)
composite_display.plot(
    "composite_reflectivity", ax=ax2, vmin=-20, vmax=80, cmap="pyart_HomeyerRainbow"
)



"""
Functions needed for reading Sigmet files from the airborne radar located on
NOAA's Hurricane Hunter aircraft.

"""

import numpy as np

from ._sigmetfile import bin2_to_angle, bin4_to_angle


def _decode_noaa_hh_hdr(
    raw_extended_headers,
    filemetadata,
    azimuth,
    elevation,
    position_source="irs",
    heading_source="irs",
):
    """
    Extract data from Sigmet extended headers produced by NOAA
    Hurricane Hunter airborne radars.

    Parameters
    ----------
    raw_extended_headers : ndarray
        Raw Sigmet extended headers.
    filemetadata : FileMetadata
        FileMetadata class from which metadata will be derived.
    azimuth : dict
        Dictionary of azimuth angles recorded in Sigmet file.
    elevation : dict
        Dictionary of elevation angles recorded in Sigmet file.
    position_source: {'irs', 'gps', 'aamps'}, optional
        Instrument from which to derive position parameters.
    heading_source: {'irs', 'aamps'}
        Instrument from which to derive heading parameters.

    Returns
    -------
    latitude : dict
        Dictionary containing latitude data and metadata.
    longitude : dict
        Dictionary containing longitude data and metadata.
    altitude : dict
        Dictionary containing altitude data and metadata.
    heading_params : dict
        Dictionary of dictionary containing aircraft heading data and
        metadata. Contains 'heading', 'roll', pitch', 'drift', 'rotation',
        'tilt' and 'georefs_applied' dictionaries.

    """
    xhdr = np.frombuffer(
        raw_extended_headers[..., :68].tostring(), dtype=list(NOAA_HH_EXTENDED_HEADER)
    )

    # rotation and tilt from azimuth/elevation angles
    rotation = filemetadata("rotation")
    tilt = filemetadata("tilt")

    rotation_data = 90.0 - elevation["data"].copy()
    rotation_data[rotation_data < 0] += 360.0
    rotation["data"] = rotation_data

    tilt_data = azimuth["data"].copy()
    tilt_data[tilt_data > 180] -= 360.0
    tilt["data"] = tilt_data

    # airborne parameters
    heading = filemetadata("heading")
    roll = filemetadata("roll")
    pitch = filemetadata("pitch")
    drift = filemetadata("drift")

    if heading_source == "irs":
        heading_data = bin2_to_angle(xhdr["irs_heading"])
        roll_data = bin2_to_angle(xhdr["irs_roll"])
        pitch_data = bin2_to_angle(xhdr["irs_pitch"])
        drift_data = bin2_to_angle(xhdr["irs_drift"])
    elif heading_source == "aamps":
        heading_data = bin2_to_angle(xhdr["aamps_heading"])
        roll_data = bin2_to_angle(xhdr["aamps_roll"])
        pitch_data = bin2_to_angle(xhdr["aamps_pitch"])
        drift_data = bin2_to_angle(xhdr["aamps_drift"])
    else:
        raise ValueError("Unknown heading_source")

    heading["data"] = heading_data
    roll["data"] = roll_data
    pitch["data"] = pitch_data
    drift["data"] = drift_data

    # georeferenced azimuth and elevation
    az, elev = _georeference_yprime(
        roll_data, pitch_data, heading_data, drift_data, rotation_data, tilt_data
    )
    azimuth["data"] = az
    elevation["data"] = elev
    georefs_applied = filemetadata("georefs_applied")
    georefs_applied["data"] = np.ones(az.shape, dtype="int8")

    # positions: latitude, longitude, altitude
    latitude = filemetadata("latitude")
    longitude = filemetadata("longitude")
    altitude = filemetadata("altitude")

    if position_source == "gps":
        lat_data = bin4_to_angle(xhdr["gps_lat"])
        lon_data = bin4_to_angle(xhdr["gps_long"])
        alt_data = xhdr["gps_alt"] / 100.0
    elif position_source == "aamps":
        lat_data = bin4_to_angle(xhdr["aamps_lat"])
        lon_data = bin4_to_angle(xhdr["aamps_long"])
        alt_data = xhdr["aamps_alt"] / 100.0
    elif position_source == "irs":
        lat_data = bin4_to_angle(xhdr["irs_lat"])
        lon_data = bin4_to_angle(xhdr["irs_long"])
        alt_data = xhdr["gps_alt"] / 100.0
    else:
        raise ValueError("Invalid position_source")

    latitude["data"] = lat_data
    longitude["data"] = lon_data
    altitude["data"] = alt_data

    extended_header_params = {
        "heading": heading,
        "roll": roll,
        "pitch": pitch,
        "drift": drift,
        "rotation": rotation,
        "tilt": tilt,
        "georefs_applied": georefs_applied,
    }
    return (latitude, longitude, altitude, extended_header_params)


def _georeference_yprime(roll, pitch, heading, drift, rotation, tilt):
    """
    Compute georeferenced azimuth and elevation angles for a Y-prime radar.

    This is the georeferencing needed for the tail doppler radar on the
    NOAA P3 aircraft.
    """
    # Adapted from Radx's SigmetRadxFile::_computeAzEl method found in
    # SigmetRadxFile.cc
    # Transforms defined in Wen-Chau Lee et al, JTech, 1994, 11, 572-578.

    # Convert to radians and use variable names from Wen-Chau Lee paper
    R = np.radians(roll)  # roll
    P = np.radians(pitch)  # pitch
    H = np.radians(heading)  # heading
    D = np.radians(drift)  # drift
    T = H + D  # track
    theta_a = np.radians(rotation)
    tau_a = np.radians(tilt)

    # Eq. (9)
    x_t = (
        np.cos(theta_a + R) * np.sin(D) * np.cos(tau_a) * np.sin(P)
        + np.cos(D) * np.sin(theta_a + R) * np.cos(tau_a)
        - np.sin(D) * np.cos(P) * np.sin(tau_a)
    )

    y_t = (
        -np.cos(theta_a + R) * np.cos(D) * np.cos(tau_a) * np.sin(P)
        + np.sin(D) * np.sin(theta_a + R) * np.cos(tau_a)
        + np.cos(P) * np.cos(D) * np.sin(tau_a)
    )

    z_t = np.cos(P) * np.cos(tau_a) * np.cos(theta_a + R) + np.sin(P) * np.sin(tau_a)

    # Eq. (12) and discussion after Eq. (17)
    lambda_t = np.arctan2(x_t, y_t)
    azimuth = np.fmod(lambda_t + T, 2 * np.pi)

    # Eq (17)
    elevation = np.arcsin(z_t)

    # convert to degrees and fix range
    azimuth = np.degrees(azimuth)
    azimuth[azimuth < 0] += 360.0
    elevation = np.degrees(elevation)
    elevation[elevation > 180] -= 360.0
    return azimuth, elevation


# NOAA Hurrican Hunter Sigmet Extended header structure
# scalar definitions
UINT16 = "H"
INT16 = "h"
BAM16 = "h"
INT32 = "i"
BAM32 = "i"

NOAA_HH_EXTENDED_HEADER = (
    ("msecs_since_sweep_start", INT32),
    ("calib_signal_level", INT16),
    ("nbytes_in_header", INT16),
    ("__pad_1", UINT16),
    ("gps_age", UINT16),  # Time in milliseconds since last GPS Input
    ("irs_age", UINT16),  # Time in milliseconds since last IRS Input
    ("aamps_age", UINT16),  # Time in milliseconds since last
    # AAMPS Input
    ("gps_lat", BAM32),  # GPS latitude (BAM)
    ("gps_long", BAM32),  # GPS Longitude (BAM)
    ("gps_alt", INT32),  # GPS Altitude (cm)
    ("gps_vel_e", INT32),  # GPS Ground Speed East (cm/second)
    ("gps_vel_n", INT32),  # GPS Ground Speed North (cm/second)
    ("gps_vel_v", INT32),  # GPS Ground Speed Up (cm/second)
    ("irs_lat", BAM32),  # IRS latitude (BAM)
    ("irs_long", BAM32),  # IRS Longitude (BAM)
    ("irs_vel_e", INT32),  # IRS Ground Speed East (cm/second)
    ("irs_vel_n", INT32),  # IRS Ground Speed North (cm/second)
    ("irs_vel_v", INT32),  # IRS Ground Speed Up (cm/second)
    ("irs_pitch", BAM16),  # IRS Pitch (BAM)
    ("irs_roll", BAM16),  # IRS Roll (BAM)
    ("irs_heading", BAM16),  # IRS Heading (BAM)
    ("irs_drift", BAM16),  # IRS Drift (BAM)
    ("irs_tru_track", BAM16),  # IRS True Track (BAM)
    ("irs_pitch_r", BAM16),  # IRS Pitch rate (BAM/sec)
    ("irs_roll_r", BAM16),  # IRS Roll rate (BAM/sec)
    ("irs_yaw_r", BAM16),  # IRS Yaw rate (BAM/sec)
    ("irs_wind_vel", INT32),  # IRS Wind speed (cm/second)
    ("irs_wind_dir", BAM16),  # IRS Wind direction (BAM)
    ("__pad_2", UINT16),
    ("aamps_lat", BAM32),  # AAMPS latitude (BAM)
    ("aamps_long", BAM32),  # AAMPS Longitude (BAM)
    ("aamps_alt", INT32),  # AAMPS Altitude (cm)
    ("aamps_ground_vel", INT32),  # AAMPS Ground Speed East (cm/second)
    ("aamps_time_stamp", INT32),  # AAMPS Timestamp in UTC
    # (seconds since the epoch)
    ("aamps_vel_v", INT32),  # AAMPS Vertical Velocity (cm/second)
    ("aamps_pitch", BAM16),  # AAMPS Pitch (BAM)
    ("aamps_roll", BAM16),  # AAMPS Roll (BAM)
    ("aamps_heading", BAM16),  # AAMPS Heading (BAM)
    ("aamps_drift", BAM16),  # AAMPS Drift (BAM)
    ("aamps_track", BAM16),  # AAMPS Track (BAM)
    ("__pad_4", UINT16),
    ("aamps_radalt_val", INT32),  # AAMPS Radar Altitude (cm)
    ("aamps_wind_vel", INT32),  # AAMPS Wind Speed (cm/second)
    ("aamps_wind_dir", BAM16),  # AAMPS Wind direction (BAM)
    ("__pad_5", UINT16),
    ("aamps_wind_vel_v", INT32),  # AAMPS Wind Speed Up (cm/second)
)



"""
GAMICFile class and utility functions.

"""

import h5py
import numpy as np


class GAMICFile:
    """
    A class to read GAMIC files.

    Parameters
    ----------
    filename : str
        Filename of GAMIC HDF5 file.

    Attributes
    ----------
    nsweeps : int
        Number of sweeps (or scans) in the file.
    rays_per_sweep : array of int32
        Number of rays in each sweep.
    total_rays : int
        Total number of rays in all sweeps.
    start_ray, end_ray : array of int32
        Index of the first (start) and last (end) ray in each sweep, 0-based.
    _hfile : HDF5 file
        Open HDF5 file object from which data is read.
    _scans : list
        Name of the HDF5 group for each scan.

    """

    def __init__(self, filename):
        """initialize object."""
        self._hfile = h5py.File(filename, "r")
        self.nsweeps = self._hfile["what"].attrs["sets"]
        self._scans = ["scan%i" % (i) for i in range(self.nsweeps)]
        self.rays_per_sweep = self.how_attrs("ray_count", "int32")
        self.total_rays = sum(self.rays_per_sweep)
        self.gates_per_sweep = self.how_attrs("bin_count", "int32")
        self.max_num_gates = max(self.gates_per_sweep)
        # check uniformity of range_step, raise if not uniform
        range_samples = self.how_attrs("range_samples", "int32")
        range_step = self.how_attrs("range_step", "float") * range_samples
        if len(np.unique(range_step)) > 1:
            raise ValueError("range scale changes between sweeps")
        # starting and ending ray for each sweep
        self.start_ray = np.cumsum(np.append([0], self.rays_per_sweep[:-1]))
        self.end_ray = np.cumsum(self.rays_per_sweep) - 1
        return

    def close(self):
        """Close the file."""
        self._hfile.close()

    # file checks
    def is_file_complete(self):
        """True if all scans in file, False otherwise."""
        # check than scan0, scan1, ... scan[nsweeps-1] are present
        for scan in self._scans:
            if scan not in self._hfile:
                return False
        return True

    def is_file_single_scan_type(self):
        """True is all scans are the same scan type, False otherwise."""
        scan_type = self._hfile["scan0/what"].attrs["scan_type"]
        for scan in self._scans:
            if self._hfile[scan]["what"].attrs["scan_type"] != scan_type:
                return False
        return True

    # attribute look up
    def where_attr(self, attr, dtype):
        """Return an array containing a attribute from the where group."""
        return np.array([self._hfile["where"].attrs[attr]], dtype=dtype)

    def how_attr(self, attr, dtype):
        """Return an array containing a attribute from the how group."""
        return np.array([self._hfile["how"].attrs[attr]], dtype=dtype)

    def is_attr_in_group(self, group, attr):
        """True is attribute is present in the group, False otherwise."""
        return attr in self._hfile[group].attrs

    def raw_group_attr(self, group, attr):
        """Return an attribute from a group with no reformatting."""
        return self._hfile[group].attrs[attr]

    def raw_scan0_group_attr(self, group, attr):
        """Return an attribute from the scan0 group with no reformatting."""
        return self._hfile["/scan0"][group].attrs[attr]

    # scan/sweep based attribute lookup
    def how_attrs(self, attr, dtype):
        """Return an array of an attribute for each scan's how group."""
        return np.array(
            [self._hfile[s]["how"].attrs[attr] for s in self._scans], dtype=dtype
        )

    def how_ext_attrs(self, attr):
        """
        Return a list of an attribute in each scan's how/extended group.
        """
        return [
            float(self._hfile[s]["how"]["extended"].attrs[attr]) for s in self._scans
        ]

    def what_attrs(self, attr, dtype):
        """Return a list of an attribute for each scan's what group."""
        return np.array(
            [self._hfile[s]["what"].attrs[attr] for s in self._scans], dtype=dtype
        )

    # misc looping
    def moment_groups(self):
        """Return a list of groups under scan0 where moments are stored."""
        return [k for k in self._hfile["/scan0"] if k.startswith("moment_")]

    def moment_names(self, scan0_groups):
        """Return a list of moment names for a list of scan0 groups."""
        if hasattr(self._hfile["/scan0"][scan0_groups[0]].attrs["moment"], "decode"):
            return [
                self._hfile["/scan0"][k].attrs["moment"].decode("utf-8")
                for k in scan0_groups
            ]
        else:
            return [self._hfile["/scan0"][k].attrs["moment"] for k in scan0_groups]

    def is_field_in_ray_header(self, field):
        """True if field is present in ray_header, False otherwise."""
        return field in self._hfile[self._scans[0]]["ray_header"].dtype.names

    def ray_header(self, field, dtype):
        """Return an array containing a ray_header field for each sweep."""
        data = np.empty((self.total_rays,), dtype=dtype)
        for scan, start, end in zip(self._scans, self.start_ray, self.end_ray):
            data[start : end + 1] = self._hfile[scan]["ray_header"][field]
        return data

    def moment_data(self, group, dtype):
        """Read in moment data from all sweeps."""
        data = np.ma.zeros((self.total_rays, self.max_num_gates), dtype=dtype)
        data[:] = np.ma.masked  # volume data initially all masked
        for scan, start, end in zip(self._scans, self.start_ray, self.end_ray):
            # read in sweep data if field exists in scan.
            if group in self._hfile[scan]:
                sweep_data = _get_gamic_sweep_data(self._hfile[scan][group])
                data[start : end + 1, : sweep_data.shape[1]] = sweep_data[:]
        return data

    def sweep_expand(self, arr, dtype="float32"):
        """Expand an sweep indexed array to be ray indexed"""
        return np.repeat(arr, self.rays_per_sweep).astype(dtype)


def _get_gamic_sweep_data(group):
    """Get GAMIC HDF5 sweep data from an HDF5 group."""
    dyn_range_min = group.attrs["dyn_range_min"]
    dyn_range_max = group.attrs["dyn_range_max"]
    raw_data = group[:]
    fmt = group.attrs["format"]
    if hasattr(fmt, "decode"):
        fmt = fmt.decode("UTF-8")
    if fmt == "UV16":
        # unsigned 16-bit integer data, 0 indicates a masked value
        assert raw_data.dtype == np.uint16
        scale = (dyn_range_max - dyn_range_min) / 65535.0
        offset = dyn_range_min
        sweep_data = np.ma.masked_array(
            raw_data * scale + offset, mask=(raw_data == 0), dtype="float32"
        )
    elif fmt == "UV8":
        # unsigned 8-bit integer data, 0 indicates a masked value
        assert raw_data.dtype == np.uint8
        scale = (dyn_range_max - dyn_range_min) / 255.0
        offset = dyn_range_min
        sweep_data = np.ma.masked_array(
            raw_data * scale + offset, mask=(raw_data == 0), dtype="float32"
        )
    elif fmt == "F":
        assert raw_data.dtype.type == np.float32
        sweep_data = np.ma.masked_array(
            raw_data, mask=np.isnan(raw_data), dtype="float32"
        )
    else:
        raise NotImplementedError("GAMIC data format: %s", fmt)
    return sweep_data



"""
======================================================
Create a multiple panel RHI plot from a CF/Radial file
======================================================

An example which creates a RHI plot of a CF/Radial file using
a RadarDisplay object.

"""
print(__doc__)

import matplotlib.pyplot as plt

import pyart
from pyart.testing import get_test_data

filename = get_test_data("sgpxsaprrhicmacI5.c0.20110524.015604_NC4.nc")

# create the plot using RadarDisplay
radar = pyart.io.read_cfradial(filename)
radar.metadata["instrument_name"] = "XSARP"
display = pyart.graph.RadarDisplay(radar)

fig = plt.figure(figsize=[15, 5])
fig.subplots_adjust(hspace=0.4)
xlabel = "Distance from radar (km)"
ylabel = "Distance above radar (km)"
colorbar_label = "Equivalent reflectivity factor (dBZ)"
# nplots = radar.nsweeps

fixed_angle = radar.fixed_angle["data"][0]
title = "HSRHI Az=%.3f" % (fixed_angle)
ax = fig.add_subplot(1, 1, 1)
display.plot(
    "reflectivity_horizontal",
    0,
    vmin=-20,
    vmax=20,
    mask_outside=True,
    title="RHI",
    axislabels=(xlabel, ylabel),
    cmap="pyart_HomeyerRainbow",
    colorbar_label=colorbar_label,
    ax=ax,
)
display.set_limits(ylim=[0, 15], ax=ax)

plt.show()



""" Unit Tests for Py-ART's correct/region_dealias.py module. """

# python test_
# to recreate a dealias_plot.png file showing the before and after doppler
# velocities

import warnings

import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal

import pyart

REF_DATA = [
    0.5,
    1.5,
    2.5,
    3.5,
    4.5,
    5.5,
    6.5,
    7.5,
    8.5,
    9.5,
    10.5,
    11.5,
    12.5,
    13.5,
    12.5,
    11.5,
    10.5,
    9.5,
    8.5,
    7.5,
    6.5,
    5.5,
    4.5,
    3.5,
    2.5,
    1.5,
    0.5,
]


def test_dealias_region_based():
    radar, dealias_vel = perform_dealias()
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False


def test_dealias_interval_limits():
    interval_limits = [-10, 0, 10]
    radar, dealias_vel = perform_dealias(interval_limits=interval_limits)
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False


def test_all_masked():
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.fields["velocity"]["data"] = np.ma.array(
        radar.fields["velocity"]["data"], mask=True
    )
    dealias_vel = pyart.correct.dealias_region_based(radar)
    assert np.all(np.ma.getmaskarray(dealias_vel["data"]))
    assert "valid_min" not in dealias_vel
    assert "valid_max" not in dealias_vel


def test_no_edges():
    # Dealiasing should work when there are no edges shared between
    # regions found
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.fields["velocity"]["data"] = np.ma.array(
        radar.fields["velocity"]["data"], mask=True
    )
    radar.fields["velocity"]["data"][0, 40] = 0.0
    radar.fields["velocity"]["data"][180, 40] = 0.0
    dealias_vel = pyart.correct.dealias_region_based(radar)
    assert_almost_equal(dealias_vel["data"][0, 40], 0.0)
    assert_almost_equal(dealias_vel["data"][180, 40], 0.0)


def test_dealias_reference_velocity_field():
    radar = pyart.testing.make_velocity_aliased_radar()
    ref_data = np.ones_like(radar.fields["velocity"]["data"]) * 20.0
    radar.add_field_like("velocity", "ref_velocity", ref_data)
    dealias_vel = pyart.correct.dealias_region_based(
        radar, ref_vel_field="ref_velocity"
    )
    offset_ref_data = np.array(REF_DATA) + 20
    assert_allclose(dealias_vel["data"][13, :27], offset_ref_data)


def test_keep_original():
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.fields["velocity"]["data"][180, 5] = 88
    gf = pyart.filters.GateFilter(radar)
    gf.exclude_above("velocity", 40)

    dealias_vel = pyart.correct.dealias_region_based(
        radar, gatefilter=gf, keep_original=False
    )
    assert np.ma.is_masked(dealias_vel["data"][180, 5]) is True

    dealias_vel = pyart.correct.dealias_region_based(
        radar, gatefilter=gf, keep_original=True
    )
    assert_almost_equal(dealias_vel["data"][180, 5], 88)
    assert np.ma.is_masked(dealias_vel["data"][180, 5]) is False


def test_set_limits():

    radar, dealias_vel = perform_dealias(set_limits=True)
    assert "valid_min" in dealias_vel
    assert_almost_equal(dealias_vel["valid_min"], -30.0)
    assert "valid_max" in dealias_vel
    assert_almost_equal(dealias_vel["valid_max"], 30.0)

    radar, dealias_vel = perform_dealias(set_limits=False)
    assert "valid_min" not in dealias_vel
    assert "valid_max" not in dealias_vel


def test_dealias_region_based_masked_wrap_around():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields["velocity"]["data"]
    radar.fields["velocity"]["data"] = np.ma.masked_array(vdata)
    radar.fields["velocity"]["data"][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields["velocity"]["data"][0, 25] = np.ma.masked
    radar.fields["velocity"]["data"][-1, 30] = np.ma.masked
    radar.fields["velocity"]["data"][180, -1] = np.ma.masked
    radar.fields["velocity"]["data"][180, 0] = np.ma.masked
    dealias_vel = pyart.correct.dealias_region_based(radar, rays_wrap_around=True)
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False
    assert np.ma.is_masked(dealias_vel["data"][0, 25])
    assert not np.ma.is_masked(dealias_vel["data"][0, 24])
    assert np.ma.is_masked(dealias_vel["data"][-1, 30])
    assert not np.ma.is_masked(dealias_vel["data"][-1, 29])
    assert np.ma.is_masked(dealias_vel["data"][180, -1])
    assert not np.ma.is_masked(dealias_vel["data"][180, -2])
    assert np.ma.is_masked(dealias_vel["data"][180, 0])
    assert not np.ma.is_masked(dealias_vel["data"][180, 1])


def test_dealias_region_based_no_wrap_around():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields["velocity"]["data"]
    radar.fields["velocity"]["data"] = np.ma.masked_array(vdata)
    radar.fields["velocity"]["data"][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields["velocity"]["data"][0, 25] = np.ma.masked
    radar.fields["velocity"]["data"][0, 25] = np.ma.masked
    radar.fields["velocity"]["data"][-1, 30] = np.ma.masked
    dealias_vel = pyart.correct.dealias_region_based(radar, rays_wrap_around=False)
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False
    assert np.ma.is_masked(dealias_vel["data"][0, 25])
    assert not np.ma.is_masked(dealias_vel["data"][0, 24])
    assert np.ma.is_masked(dealias_vel["data"][-1, 30])
    assert not np.ma.is_masked(dealias_vel["data"][-1, 29])


def perform_dealias(**kwargs):
    """Perform velocity dealiasing on reference data."""
    radar = pyart.testing.make_velocity_aliased_radar()
    # speckling that will not be not be dealiased.
    radar.fields["velocity"]["data"][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_region_based(radar, **kwargs)
    return radar, dealias_vel


def test_dealias_outside():
    # unwrap when the velocity field contains gates with values outside of
    # the interval limit, but a warnings should be raised
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.fields["velocity"]["data"][250, 25] = 20.1
    radar.fields["velocity"]["data"][270, 25] = -20.1
    with warnings.catch_warnings(record=True) as w:
        dealias_vel = pyart.correct.dealias_region_based(radar)
        assert len(w) == 1
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False
    assert_almost_equal(dealias_vel["data"][250, 25], 0.10, 2)
    assert_almost_equal(dealias_vel["data"][270, 25], -0.10, 2)


def test_fillvalue(**kwargs):
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.fields["velocity"]["_FillValue"] = 8888.0

    # ensure that at least one gate is excluded and should be masked
    radar.fields["velocity"]["data"][100, 25] = 44
    gf = pyart.filters.GateFilter(radar)
    gf.exclude_above("velocity", 40)

    dealias_vel = pyart.correct.dealias_region_based(radar, gatefilter=gf)
    assert isinstance(dealias_vel["data"], np.ma.MaskedArray)
    assert np.ma.is_masked(dealias_vel["data"][100, 25])

    assert "_FillValue" in dealias_vel
    assert dealias_vel["_FillValue"] == 8888.0
    assert dealias_vel["data"].fill_value == 8888.0


def main():

    radar, dealias_vel = perform_dealias()
    radar.fields["dealiased_velocity"] = dealias_vel

    # print out results
    print("ray 13 velocitites before dealias:")
    print(radar.fields["velocity"]["data"][13])
    print("ray 13 velocities after dealias:")
    print(radar.fields["dealiased_velocity"]["data"][13])

    # create plot
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=[5, 10])
    ax1 = fig.add_subplot(2, 1, 1)
    ax2 = fig.add_subplot(2, 1, 2)
    rd = pyart.graph.RadarDisplay(radar)
    rd.plot_ppi("velocity", 0, ax=ax1, colorbar_flag=False, title="", vmin=-10, vmax=10)
    rd.plot_ppi(
        "dealiased_velocity",
        0,
        ax=ax2,
        colorbar_flag=False,
        title="",
        vmin=-10,
        vmax=20,
    )
    plt.show()
    fig.savefig("dealias_plot.png")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Python ARM Radar Toolkit documentation build configuration file, created by
# sphinx-quickstart on Thu Jun 28 12:35:56 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
import os
import re

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
    "matplotlib.sphinxext.plot_directive",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinx_gallery.gen_gallery",
    "sphinx_gallery.load_style",
    "nbsphinx",
    "myst_nb",
    "ablog",
]

exclude_patterns = ["_build", "**.ipynb_checkpoints"]

# only include examples if the BUILD_PYART_EXAMPLES env. variable is set
extensions.append("sphinx_gallery.gen_gallery")
sphinx_gallery_conf = {"examples_dirs": "../../examples", "gallery_dirs": "examples"}


# Configuration options for plot_directive. See:
# https://github.com/matplotlib/matplotlib/blob/f3ed922d935751e08494e5fb5311d3050a3b637b/lib/matplotlib/sphinxext/plot_directive.py#L81
plot_html_show_source_link = False
plot_html_show_formats = False
plot_formats = ["png"]
plot_rcparams = {"savefig.bbox": "tight"}

# Generate the API documentation when building
autoclass_content = "both"

autosummary_generate = True
autosummary_imported_members = True

# Otherwise, the Return parameter list looks different from the Parameters list
napoleon_use_rtype = False
# Otherwise, the Attributes parameter list looks different from the Parameters list
napoleon_use_ivar = True
napoleon_include_init_with_doc = False
napoleon_use_param = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst", ".md", ".ipynb"]

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Py-ART"
copyright = "2013-2022, Py-ART developers"
author = "Py-ART developers"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
import pyart

# The short X.Y version (including the .devXXXX suffix if present)
version = re.sub(r"^(\d+\.\d+)\.\d+(.*)", r"\1\2", pyart.__version__)
if "dev" in version:
    # retain the .dev suffix, but clean it up
    version = re.sub(r"(\.dev\d*).*?$", r"\1", version)
else:
    # strip all other suffixes
    version = re.sub(r"^(\d+\.\d+).*?$", r"\1", version)
# The full version, including alpha/beta/rc tags.
release = pyart.__version__
# full Py-ART version in CI built docs
if "CI" in os.environ and os.environ["CI"] == "true":
    version = release
# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["*.ipynb"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = "pydata_sphinx_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "google_analytics_id": "G-JJEG3CV376",
    "github_url": "https://github.com/ARM-DOE/pyart",
    "twitter_url": "https://twitter.com/Py_ART",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_css_files = ["pyart-theme.css"]

html_js_files = ["doc_shared.js"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "userguide/*": ["searchbox.html", "sidebar-nav-bs.html"],
    "API/*": ["searchbox.html", "sidebar-nav-bs.html"],
    "examples/*": ["searchbox.html", "sidebar-nav-bs.html"],
    "notebook-gallery": ["searchbox.html", "sidebar-nav-bs.html"],
    "changelog": ["searchbox.html", "sidebar-nav-bs.html"],
    "blog": [
        "searchbox.html",
        "sidebar-nav-bs.html",
        "tagcloud.html",
        "recentposts.html",
        "archives.html",
    ],
    "blog_posts/*/*": [
        "searchbox.html",
        "sidebar-nav-bs.html",
        "postcard.html",
        "recentposts.html",
        "archives.html",
    ],
}

# Setup the blog portion
blog_baseurl = "mgrover1.github.io/pyart/"
blog_title = "PyART Blog"
blog_path = "blog"
fontawesome_included = True
blog_post_pattern = "blog_posts/*/*"
post_redirect_refresh = 1
post_auto_image = 1
post_auto_excerpt = 2

# Don't execute the jupyter notebooks
jupyter_execute_notebooks = "off"

# Extra variables that will be available to the templates. Used to create the
# links to the Github repository sources and issues
html_context = {
    "doc_path": "doc",
    "galleries": sphinx_gallery_conf["gallery_dirs"],
    "gallery_dir": dict(
        zip(sphinx_gallery_conf["gallery_dirs"], sphinx_gallery_conf["examples_dirs"])
    ),
    "api_dir": "API/generated",
    "github_user": "ARM-DOE",
    "github_repo": "pyart",
    "github_version": "main",
}

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "pyart"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "pyart.tex", "Py-ART Documentation", "Contributors", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "pyart", "The Python ARM Radar Toolkit", [author], 1)]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "matplotlib": ("https://matplotlib.org", None),
}

# Add myst extensions
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "html_image",
    "dollarmath",
]



""" Unit Tests for Py-ART's util/radar_utils.py module. """

import numpy as np

import pyart


def test_is_vpt():
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    assert not pyart.util.is_vpt(radar)
    pyart.util.to_vpt(radar)
    assert pyart.util.is_vpt(radar)


def test_join_radar():
    radar1 = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    radar1.instrument_parameters = {"nyquist_velocity": {"data": np.array([6] * 3)}}
    field = {"data": np.ones((36 * 3, 10))}
    radar1.add_field("f1", field)
    radar1.add_field("f2", field)
    radar2 = pyart.testing.make_empty_ppi_radar(10, 36, 4)
    radar2.instrument_parameters = {"nyquist_velocity": {"data": np.array([8] * 4)}}
    field = {"data": np.ones((36 * 4, 10))}
    radar2.add_field("f1", field)

    radar3 = pyart.util.join_radar(radar1, radar2)
    assert "f1" in radar3.fields
    assert len(radar3.instrument_parameters["nyquist_velocity"]["data"]) == 7


def test_to_vpt():
    # single scan
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    radar.instrument_parameters = {"prt_mode": {"data": np.array(["fixed"] * 3)}}
    pyart.util.to_vpt(radar)
    assert pyart.util.is_vpt(radar)
    assert radar.nsweeps == 1
    assert radar.azimuth["data"][10] == 0.0
    assert radar.elevation["data"][0] == 90.0
    assert len(radar.instrument_parameters["prt_mode"]["data"]) == 1

    # multiple scans
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    radar.instrument_parameters = {"prt_mode": {"data": np.array(["fixed"] * 3)}}
    pyart.util.to_vpt(radar, False)
    assert pyart.util.is_vpt(radar)
    assert radar.nsweeps == 108
    assert radar.azimuth["data"][10] == 10.0
    assert radar.elevation["data"][0] == 90.0
    assert len(radar.instrument_parameters["prt_mode"]["data"]) == 108


def test_subset_radar():
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    field = {"data": np.ones((36 * 3, 10))}
    radar.add_field("f1", field)
    radar.add_field("f2", field)
    azi_min = 10
    azi_max = 100
    rng_min = 200
    rng_max = 800
    ele_min = 0.75
    ele_max = 0.75
    radarcut = pyart.util.radar_utils.subset_radar(
        radar,
        ["f1"],
        rng_min=200,
        rng_max=800,
        ele_min=0.75,
        ele_max=0.75,
        azi_min=10,
        azi_max=100,
    )
    # assert correct domain and correct fields
    assert radarcut.azimuth["data"].min() >= azi_min
    assert radarcut.azimuth["data"].max() <= azi_max
    assert radarcut.range["data"].min() >= rng_min
    assert radarcut.range["data"].max() <= rng_max
    assert radarcut.elevation["data"].min() >= ele_min
    assert radarcut.elevation["data"].max() <= ele_max
    assert list(radarcut.fields) == ["f1"]


# read in example file
radar = pyart.io.read_nexrad_archive(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE)


def test_image_mute_radar():

    # image mute example file
    mute_radar = pyart.util.image_mute_radar(
        radar,
        field="reflectivity",
        mute_field="cross_correlation_ratio",
        mute_threshold=0.97,
    )

    # check that fields are added to the radar object
    assert "nonmuted_reflectivity" in mute_radar.fields.keys()
    assert "muted_reflectivity" in mute_radar.fields.keys()

    # check that the number of points in the muted and non muted reflectivity
    # fields have the same number of points as the rhoHV field
    n_rhohv = np.sum(~mute_radar.fields["cross_correlation_ratio"]["data"].mask)
    n_mutez = np.sum(~mute_radar.fields["muted_reflectivity"]["data"].mask)
    n_nonmutez = np.sum(~mute_radar.fields["nonmuted_reflectivity"]["data"].mask)

    assert n_mutez + n_nonmutez == n_rhohv



"""
Functions for reading NEXRAD Level 3 products.

"""

import numpy as np

from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from .common import _test_arguments, make_time_unit_str, prepare_for_read
from .nexrad_level3 import NEXRADLevel3File


def read_nexrad_level3(
    filename,
    field_names=None,
    additional_metadata=None,
    file_field_names=False,
    exclude_fields=None,
    include_fields=None,
    storage_options={"anon": True},
    **kwargs
):
    """
    Read a NEXRAD Level 3 product.

    Parameters
    ----------
    filename : str
        Filename of NEXRAD Level 3 product file. The files hosted by
        at the NOAA National Climate Data Center [1]_ as well as on the
        NWS WSR-88D Level III Data Collection and Distribution Network
        have been tests. Other NEXRAD Level 3 files may or may not work.
        A file-like object pointing to the beginning of such a file is also
        supported [2]_.
    field_names : dict, optional
        Dictionary mapping NEXRAD level 3 product number to radar field names.
        If the product number of the file does not appear in this dictionary
        or has a value of None it will not be placed in the radar.fields
        dictionary. A value of None, the default, will use the mapping
        defined in the metadata configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included. A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the metadata configuration file will be used.
    file_field_names : bool, optional
        True to use the product number for the field name. In this case the
        field_names parameter is ignored. The field dictionary will likely
        only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    **kwargs
        Additional keyword arguments to pass to fsspec to open the dataset

    Returns
    -------
    radar : Radar
        Radar object containing all moments and sweeps/cuts in the volume.
        Gates not collected are masked in the field data.

    References
    ----------
    .. [1] http://www.ncdc.noaa.gov/
    .. [2] http://www.roc.noaa.gov/wsr88d/Level_III/Level3Info.asp

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata(
        "nexrad_level3",
        field_names,
        additional_metadata,
        file_field_names,
        exclude_fields,
        include_fields,
    )

    # open the file
    nfile = NEXRADLevel3File(
        prepare_for_read(filename, storage_options=storage_options)
    )
    nradials = nfile.packet_header["nradials"]
    msg_code = nfile.msg_header["code"]

    # time
    time = filemetadata("time")
    time_start = nfile.get_volume_start_datetime()
    time["units"] = make_time_unit_str(time_start)
    time["data"] = np.zeros((nradials,), dtype="float64")

    # range
    _range = filemetadata("range")
    _range["data"] = nfile.get_range()
    _range["meters_to_center_of_first_gate"] = _range["data"][0]
    _range["meters_between_gates"] = _range["data"][1] - _range["data"][0]

    # fields
    fields = {}
    field_name = filemetadata.get_field_name(msg_code)
    if field_name is None:
        fields = {}
    else:
        dic = filemetadata(field_name)
        dic["_FillValue"] = get_fillvalue()
        dic["data"] = nfile.get_data()
        fields = {field_name: dic}

    # metadata
    metadata = filemetadata("metadata")
    metadata["original_container"] = "NEXRAD Level 3"

    # scan_type
    scan_type = "ppi"

    # latitude, longitude, altitude
    latitude = filemetadata("latitude")
    longitude = filemetadata("longitude")
    altitude = filemetadata("altitude")

    lat, lon, height = nfile.get_location()
    # Nexrad altitude is in feet, convert to meters unless user's
    # default config has units in feet.
    if altitude["units"] == "meters":
        height = height * 0.3048
    latitude["data"] = np.array([lat], dtype="float64")
    longitude["data"] = np.array([lon], dtype="float64")
    altitude["data"] = np.array([height], dtype="float64")

    # sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index
    # sweep_end_ray_index
    sweep_number = filemetadata("sweep_number")
    sweep_mode = filemetadata("sweep_mode")
    sweep_start_ray_index = filemetadata("sweep_start_ray_index")
    sweep_end_ray_index = filemetadata("sweep_end_ray_index")

    sweep_number["data"] = np.array([0], dtype="int32")
    sweep_mode["data"] = np.array(1 * ["azimuth_surveillance"], dtype="S")

    sweep_start_ray_index["data"] = np.array([0], dtype="int32")
    sweep_end_ray_index["data"] = np.array([nradials - 1], dtype="int32")

    # azimuth, elevation, fixed_angle
    azimuth = filemetadata("azimuth")
    elevation = filemetadata("elevation")
    fixed_angle = filemetadata("fixed_angle")
    azimuth["data"] = nfile.get_azimuth()
    elev = nfile.get_elevation()
    elevation["data"] = np.ones((nradials,), dtype="float32") * elev
    fixed_angle["data"] = np.array([elev], dtype="float32")

    nfile.close()
    return Radar(
        time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        instrument_parameters=None,
    )



"""
Core Py-ART classes and function for interacting with weather radar data.

"""

from .grid import Grid  # noqa
from .radar import Radar  # noqa
from .radar_spectra import RadarSpectra  # noqa
from .transforms import antenna_to_cartesian  # noqa
from .transforms import antenna_vectors_to_cartesian  # noqa
from .transforms import cartesian_to_geographic  # noqa
from .transforms import cartesian_to_geographic_aeqd  # noqa
from .transforms import cartesian_vectors_to_geographic  # noqa
from .transforms import geographic_to_cartesian  # noqa
from .transforms import geographic_to_cartesian_aeqd  # noqa
from .wind_profile import HorizontalWindProfile  # noqa

__all__ = [s for s in dir() if not s.startswith("_")]



"""
Functions for working radar instances.

"""

import copy

import numpy as np
from netCDF4 import date2num

from ..config import get_fillvalue
from . import datetime_utils


def is_vpt(radar, offset=0.5):
    """
    Determine if a Radar appears to be a vertical pointing scan.

    This function only verifies that the object is a vertical pointing scan,
    use the :py:func:`to_vpt` function to convert the radar to a vpt scan
    if this function returns True.

    Parameters
    ----------
    radar : Radar
        Radar object to determine if.
    offset : float, optional
        Maximum offset of the elevation from 90 degrees to still consider
        to be vertically pointing.

    Returns
    -------
    flag : bool
        True if the radar appear to be verticle pointing, False if not.

    """
    # check that the elevation is within offset of 90 degrees.
    elev = radar.elevation["data"]
    return np.all((elev < 90.0 + offset) & (elev > 90.0 - offset))


def to_vpt(radar, single_scan=True):
    """
    Convert an existing Radar object to represent a vertical pointing scan.

    This function does not verify that the Radar object contains a vertical
    pointing scan. To perform such a check use :py:func:`is_vpt`.

    Parameters
    ----------
    radar : Radar
        Mislabeled vertical pointing scan Radar object to convert to be
        properly labeled. This object is converted in place, no copy of
        the existing data is made.
    single_scan : bool, optional
        True to convert the volume to a single scan, any azimuth angle data
        is lost. False will convert the scan to contain the same number of
        scans as rays, azimuth angles are retained.

    """
    if single_scan:
        nsweeps = 1
        radar.azimuth["data"][:] = 0.0
        seri = np.array([radar.nrays - 1], dtype="int32")
        radar.sweep_end_ray_index["data"] = seri
    else:
        nsweeps = radar.nrays
        # radar.azimuth not adjusted
        radar.sweep_end_ray_index["data"] = np.arange(nsweeps, dtype="int32")

    radar.scan_type = "vpt"
    radar.nsweeps = nsweeps
    radar.target_scan_rate = None  # no scanning
    radar.elevation["data"][:] = 90.0

    radar.sweep_number["data"] = np.arange(nsweeps, dtype="int32")
    radar.sweep_mode["data"] = np.array(["vertical_pointing"] * nsweeps)
    radar.fixed_angle["data"] = np.ones(nsweeps, dtype="float32") * 90.0
    radar.sweep_start_ray_index["data"] = np.arange(nsweeps, dtype="int32")

    if radar.instrument_parameters is not None:
        for key in ["prt_mode", "follow_mode", "polarization_mode"]:
            if key in radar.instrument_parameters:
                ip_dic = radar.instrument_parameters[key]
                ip_dic["data"] = np.array([ip_dic["data"][0]] * nsweeps)

    # Attributes that do not need any changes
    # radar.altitude
    # radar.altitude_agl
    # radar.latitude
    # radar.longitude

    # radar.range
    # radar.ngates
    # radar.nrays

    # radar.metadata
    # radar.radar_calibration

    # radar.time
    # radar.fields
    # radar.antenna_transition
    # radar.scan_rate
    return


def subset_radar(
    radar,
    field_names,
    rng_min=None,
    rng_max=None,
    ele_min=None,
    ele_max=None,
    azi_min=None,
    azi_max=None,
):
    """
    Creates a subset of the radar object along new dimensions.

    Parameters
    ----------
    radar : radar object
        The radar object containing the data.
    field_names : str or None
        The fields to keep in the new radar.
    rng_min, rng_max : float
        The range limits [m]. If None the entire coverage of the radar is
        going to be used.
    ele_min, ele_max, azi_min, azi_max : float or None
        The limits of the grid [deg]. If None the limits will be the limits
        of the radar volume.

    Returns
    -------
    radar : radar object
        The radar object containing only the desired data.

    """
    radar_aux = copy.deepcopy(radar)

    if (
        rng_min is None
        and rng_max is None
        and ele_min is None
        and ele_max is None
        and azi_min is None
        and azi_max is None
    ):
        return radar_aux

    if rng_min is None:
        rng_min = 0.0
    if rng_max is None:
        rng_max = np.max(radar_aux.range["data"])

    ind_rng = np.where(
        np.logical_and(
            radar_aux.range["data"] >= rng_min, radar_aux.range["data"] <= rng_max
        )
    )[0]

    if ind_rng.size == 0:
        print("No range bins between " + str(rng_min) + " and " + str(rng_max) + " m")
        return None

    # Determine angle limits
    if radar_aux.scan_type == "ppi":
        if ele_min is None:
            ele_min = np.min(radar_aux.fixed_angle["data"])
        if ele_max is None:
            ele_max = np.max(radar_aux.fixed_angle["data"])
        if azi_min is None:
            azi_min = np.min(radar_aux.azimuth["data"])
        if azi_max is None:
            azi_max = np.max(radar_aux.azimuth["data"])
    else:
        if ele_min is None:
            ele_min = np.min(radar_aux.elevation["data"])
        if ele_max is None:
            ele_max = np.max(radar_aux.elevation["data"])
        if azi_min is None:
            azi_min = np.min(radar_aux.fixed_angle["data"])
        if azi_max is None:
            azi_max = np.max(radar_aux.fixed_angle["data"])

    if radar_aux.scan_type == "ppi":
        # Get radar elevation angles within limits
        ele_vec = np.sort(radar_aux.fixed_angle["data"])
        ele_vec = ele_vec[np.logical_and(ele_vec >= ele_min, ele_vec <= ele_max)]
        if ele_vec.size == 0:
            print(
                "No elevation angles between " + str(ele_min) + " and " + str(ele_max)
            )
            return None

        # get sweeps corresponding to the desired elevation angles
        ind_sweeps = []
        for ele in ele_vec:
            ind_sweeps.append(np.where(radar_aux.fixed_angle["data"] == ele)[0][0])
        radar_aux = radar_aux.extract_sweeps(ind_sweeps)

        # Get indices of rays within limits
        if azi_min < azi_max:
            ind_rays = np.where(
                np.logical_and(
                    radar_aux.azimuth["data"] >= azi_min,
                    radar_aux.azimuth["data"] <= azi_max,
                )
            )[0]
        else:
            ind_rays = np.where(
                np.logical_or(
                    radar_aux.azimuth["data"] >= azi_min,
                    radar_aux.azimuth["data"] <= azi_max,
                )
            )[0]

    else:
        # Get radar azimuth angles within limits
        azi_vec = radar_aux.fixed_angle["data"]
        if azi_min < azi_max:
            azi_vec = np.sort(
                azi_vec[np.logical_and(azi_vec >= azi_min, azi_vec <= azi_max)]
            )
        else:
            azi_vec = azi_vec[np.logical_or(azi_vec >= azi_min, azi_vec <= azi_max)]
        if azi_vec.size == 0:
            print("No azimuth angles between " + str(azi_min) + " and " + str(azi_max))
            return None

        # get sweeps corresponding to the desired azimuth angles
        ind_sweeps = []
        for azi in azi_vec:
            ind_sweeps.append(np.where(radar_aux.fixed_angle["data"] == azi)[0][0])
        radar_aux = radar_aux.extract_sweeps(ind_sweeps)

        # Get indices of rays within limits
        ind_rays = np.where(
            np.logical_and(
                radar_aux.elevation["data"] >= ele_min,
                radar_aux.elevation["data"] <= ele_max,
            )
        )[0]

    # get new sweep start index and stop index
    sweep_start_inds = copy.deepcopy(radar_aux.sweep_start_ray_index["data"])
    sweep_end_inds = copy.deepcopy(radar_aux.sweep_end_ray_index["data"])

    nrays = 0
    ind_rays_aux = []
    for j in range(radar_aux.nsweeps):
        # get ray indices for this sweep
        ind_rays_sweep = ind_rays[
            np.logical_and(
                ind_rays >= sweep_start_inds[j], ind_rays <= sweep_end_inds[j]
            )
        ]
        # order rays
        if radar_aux.scan_type == "ppi":
            ind = np.argsort(radar_aux.azimuth["data"][ind_rays_sweep])
            ind_rays_sweep = ind_rays_sweep[ind]
            # avoid large gaps in data
            azimuths = radar_aux.azimuth["data"][ind_rays_sweep]
            azi_steps = azimuths[1:] - azimuths[:-1]
            ind_gap = np.where(azi_steps > 2 * np.median(azi_steps))[0]
            if ind_gap.size > 0:
                ind_rays_sweep = np.append(
                    ind_rays_sweep[ind_gap[0] + 1 :], ind_rays_sweep[: ind_gap[0] + 1]
                )
            ind_rays_aux.extend(ind_rays_sweep)
        else:
            ind = np.argsort(radar_aux.elevation["data"][ind_rays_sweep])
            ind_rays_aux.extend(ind_rays_sweep[ind])

        rays_in_sweep = ind_rays_sweep.size
        radar_aux.rays_per_sweep["data"][j] = rays_in_sweep
        if j == 0:
            radar_aux.sweep_start_ray_index["data"][j] = 0
        else:
            radar_aux.sweep_start_ray_index["data"][j] = int(
                radar_aux.sweep_end_ray_index["data"][j - 1] + 1
            )
        radar_aux.sweep_end_ray_index["data"][j] = (
            radar_aux.sweep_start_ray_index["data"][j] + rays_in_sweep - 1
        )
        nrays += rays_in_sweep

    ind_rays = np.array(ind_rays_aux)

    # Update metadata
    radar_aux.range["data"] = radar_aux.range["data"][ind_rng]
    radar_aux.time["data"] = radar_aux.time["data"][ind_rays]
    radar_aux.azimuth["data"] = radar_aux.azimuth["data"][ind_rays]
    radar_aux.elevation["data"] = radar_aux.elevation["data"][ind_rays]
    radar_aux.init_gate_x_y_z()
    radar_aux.init_gate_longitude_latitude()
    radar_aux.init_gate_altitude()
    radar_aux.nrays = nrays
    radar_aux.ngates = ind_rng.size

    if radar_aux.instrument_parameters is not None:
        if "nyquist_velocity" in radar_aux.instrument_parameters:
            radar_aux.instrument_parameters["nyquist_velocity"][
                "data"
            ] = radar_aux.instrument_parameters["nyquist_velocity"]["data"][ind_rays]
        if "pulse_width" in radar_aux.instrument_parameters:
            radar_aux.instrument_parameters["pulse_width"][
                "data"
            ] = radar_aux.instrument_parameters["pulse_width"]["data"][ind_rays]
        if "number_of_pulses" in radar_aux.instrument_parameters:
            radar_aux.instrument_parameters["number_of_pulses"][
                "data"
            ] = radar_aux.instrument_parameters["number_of_pulses"]["data"][ind_rays]

    # Get new fields
    if field_names is None:
        radar_aux.fields = dict()
    else:
        fields_aux = copy.deepcopy(radar_aux.fields)
        radar_aux.fields = dict()
        for field_name in field_names:
            if field_name not in fields_aux:
                print("Field " + field_name + " not available")
                continue

            fields_aux[field_name]["data"] = fields_aux[field_name]["data"][:, ind_rng]
            fields_aux[field_name]["data"] = fields_aux[field_name]["data"][ind_rays, :]
            radar_aux.add_field(field_name, fields_aux[field_name])

    return radar_aux


def join_radar(radar1, radar2):
    """
    Combine two radar instances into one.

    Parameters
    ----------
    radar1 : Radar
        Radar object.
    radar2 : Radar
        Radar object.

    """
    # must have same gate spacing
    new_radar = copy.deepcopy(radar1)
    new_radar.azimuth["data"] = np.append(
        radar1.azimuth["data"], radar2.azimuth["data"]
    )
    new_radar.elevation["data"] = np.append(
        radar1.elevation["data"], radar2.elevation["data"]
    )
    new_radar.fixed_angle["data"] = np.append(
        radar1.fixed_angle["data"], radar2.fixed_angle["data"]
    )
    new_radar.sweep_number["data"] = np.append(
        radar1.sweep_number["data"], radar2.sweep_number["data"]
    )
    new_radar.sweep_start_ray_index["data"] = np.append(
        radar1.sweep_start_ray_index["data"],
        radar2.sweep_start_ray_index["data"] + radar1.nrays,
    )
    new_radar.sweep_end_ray_index["data"] = np.append(
        radar1.sweep_end_ray_index["data"],
        radar2.sweep_end_ray_index["data"] + radar1.nrays,
    )
    new_radar.nsweeps += radar2.nsweeps
    new_radar.sweep_mode["data"] = np.append(
        radar1.sweep_mode["data"], radar2.sweep_mode["data"]
    )
    if (radar1.rays_are_indexed is not None) and (radar2.rays_are_indexed is not None):
        new_radar.rays_are_indexed["data"] = np.append(
            radar1.rays_are_indexed["data"], radar2.rays_are_indexed["data"]
        )
    else:
        new_radar.rays_are_indexed = None

    if new_radar.instrument_parameters is not None:
        if "nyquist_velocity" in new_radar.instrument_parameters:
            new_radar.instrument_parameters["nyquist_velocity"]["data"] = np.append(
                radar1.instrument_parameters["nyquist_velocity"]["data"],
                radar2.instrument_parameters["nyquist_velocity"]["data"],
            )
        if "pulse_width" in new_radar.instrument_parameters:
            new_radar.instrument_parameters["pulse_width"]["data"] = np.append(
                radar1.instrument_parameters["pulse_width"]["data"],
                radar2.instrument_parameters["pulse_width"]["data"],
            )
        if "number_of_pulses" in new_radar.instrument_parameters:
            new_radar.instrument_parameters["number_of_pulses"]["data"] = np.append(
                radar1.instrument_parameters["number_of_pulses"]["data"],
                radar2.instrument_parameters["number_of_pulses"]["data"],
            )
        if "prt" in new_radar.instrument_parameters:
            new_radar.instrument_parameters["prt"]["data"] = np.append(
                radar1.instrument_parameters["prt"]["data"],
                radar2.instrument_parameters["prt"]["data"],
            )

    if (radar1.ray_angle_res is not None) and (radar2.ray_angle_res is not None):
        new_radar.ray_angle_res["data"] = np.append(
            radar1.ray_angle_res["data"], radar2.ray_angle_res["data"]
        )
    else:
        new_radar.ray_angle_res = None

    if len(radar1.range["data"]) >= len(radar2.range["data"]):
        new_radar.range["data"] = radar1.range["data"]
    else:
        new_radar.range["data"] = radar2.range["data"]
    new_radar.ngates = len(new_radar.range["data"])

    if (radar1.target_scan_rate is not None) and (radar2.target_scan_rate is not None):
        new_radar.target_scan_rate["data"] = np.append(
            radar1.target_scan_rate["data"], radar2.target_scan_rate["data"]
        )
    else:
        new_radar.target_scan_rate = None

    # to combine times we need to reference them to a standard
    # for this we'll use epoch time
    r1num = datetime_utils.datetimes_from_radar(radar1, epoch=True)
    r2num = datetime_utils.datetimes_from_radar(radar2, epoch=True)
    new_radar.time["data"] = date2num(
        np.append(r1num, r2num), datetime_utils.EPOCH_UNITS
    )
    new_radar.time["units"] = datetime_utils.EPOCH_UNITS
    new_radar.nrays = len(new_radar.time["data"])

    fields_to_remove = []
    for var in new_radar.fields.keys():
        # if the field is present in both radars combine both fields
        # otherwise remove it from new radar
        if var in radar1.fields and var in radar2.fields:
            sh1 = radar1.fields[var]["data"].shape
            sh2 = radar2.fields[var]["data"].shape
            new_field_shape = (sh1[0] + sh2[0], max(sh1[1], sh2[1]))
            new_field = np.ma.masked_all(new_field_shape)
            new_field.set_fill_value(get_fillvalue())
            new_field[0 : sh1[0], 0 : sh1[1]] = radar1.fields[var]["data"]
            new_field[sh1[0] :, 0 : sh2[1]] = radar2.fields[var]["data"]
            new_radar.fields[var]["data"] = new_field
        else:
            print("Field " + var + " not present in both radars")
            fields_to_remove.append(var)

    if fields_to_remove:
        for field_name in fields_to_remove:
            new_radar.fields.pop(field_name, None)

    # radar locations
    # TODO moving platforms - any more?
    if (
        len(radar1.latitude["data"])
        == 1 & len(radar2.latitude["data"])
        == 1 & len(radar1.longitude["data"])
        == 1 & len(radar2.longitude["data"])
        == 1 & len(radar1.altitude["data"])
        == 1 & len(radar2.altitude["data"])
        == 1
    ):

        lat1 = float(radar1.latitude["data"])
        lon1 = float(radar1.longitude["data"])
        alt1 = float(radar1.altitude["data"])
        lat2 = float(radar2.latitude["data"])
        lon2 = float(radar2.longitude["data"])
        alt2 = float(radar2.altitude["data"])

        if (lat1 != lat2) or (lon1 != lon2) or (alt1 != alt2):
            ones1 = np.ones(len(radar1.time["data"]), dtype="float32")
            ones2 = np.ones(len(radar2.time["data"]), dtype="float32")
            new_radar.latitude["data"] = np.append(ones1 * lat1, ones2 * lat2)
            new_radar.longitude["data"] = np.append(ones1 * lon1, ones2 * lon2)
            new_radar.latitude["data"] = np.append(ones1 * alt1, ones2 * alt2)
        else:
            new_radar.latitude["data"] = radar1.latitude["data"]
            new_radar.longitude["data"] = radar1.longitude["data"]
            new_radar.altitude["data"] = radar1.altitude["data"]

    else:
        new_radar.latitude["data"] = np.append(
            radar1.latitude["data"], radar2.latitude["data"]
        )
        new_radar.longitude["data"] = np.append(
            radar1.longitude["data"], radar2.longitude["data"]
        )
        new_radar.altitude["data"] = np.append(
            radar1.altitude["data"], radar2.altitude["data"]
        )

    return new_radar


def image_mute_radar(radar, field, mute_field, mute_threshold, field_threshold=None):
    """
    This function will split a field based on thresholds from another field.

    Specifically, it was designed to separate areas of reflectivity where
    the correlation coefficient is less than a certain threshold to discern
    melting precipitation.

    Parameters
    ----------
    radar : Radar
        Radar instance which provides the fields for muting.
    field : str
        Name of field to image mute.
    mute_field : str
        Name of field to image mute by.
    mute_threshold : float
        Threshold value to mute by.
    field_threshold : float
        Additional threshold to mask.

    Returns
    -------
    radar : Radar
        Radar object with 2 new fields from input field, one muted and one not muted.

    References
    ----------
    Tomkins, L. M., Yuter, S. E., Miller, M. A., and Allen, L. R., 2022:
    Image muting of mixed precipitation to improve identification of regions
    of heavy snow in radar data. Atmos. Meas. Tech., 15, 5515–5525,
    https://doi.org/10.5194/amt-15-5515-2022

    """
    # add checks for field availability
    if field not in radar.fields.keys():
        raise KeyError("Failed - ", field, " field to mute not found in Radar object.")

    if mute_field not in radar.fields.keys():
        raise KeyError(
            "Failed - ", mute_field, " field to mute by not found in Radar object."
        )

    # get data from fields
    data_to_mute = radar.fields[field]["data"]
    data_mute_by = radar.fields[mute_field]["data"]

    # create filters
    # field_filter is used if user wants to use additional criteria in the
    # original field
    if field_threshold is not None:
        field_filter = data_to_mute >= field_threshold
    else:
        field_filter = None

    # mute_filter will be the primary filter for determining muted regions
    mute_filter = data_mute_by <= mute_threshold

    # mute_mask is the combined filter
    if field_filter is None:
        mute_mask = mute_filter
    else:
        mute_mask = mute_filter & field_filter

    # break up the field into muted regions and non muted regions
    non_muted_field = np.ma.masked_where(mute_mask, data_to_mute)
    non_muted_field = np.ma.masked_invalid(non_muted_field)

    muted_field = np.ma.masked_where(~mute_mask, data_to_mute)
    muted_field = np.ma.masked_invalid(muted_field)

    # add fields to a dictionary and save to radar object
    non_muted_dict = radar.fields[field].copy()
    non_muted_dict["data"] = non_muted_field
    non_muted_dict["long_name"] = "Non-muted " + field
    radar.add_field("nonmuted_" + field, non_muted_dict)

    muted_dict = radar.fields[field].copy()
    muted_dict["data"] = muted_field
    muted_dict["long_name"] = "Muted " + field
    radar.add_field("muted_" + field, muted_dict)
    return radar



""" Unit tests for the hildebrand_sekhon module. """

import numpy as np
from numpy.testing import assert_almost_equal

from pyart.util import hildebrand_sekhon

SAMPLE_SPECTRUM = np.array(
    [
        11.089,
        10.386,
        11.081,
        9.696,
        9.21,
        11.103,
        9.425,
        10.222,
        8.393,
        11.67,
        9.632,
        10.096,
        9.807,
        10.374,
        8.065,
        8.682,
        10.164,
        320.206,
        480.898,
        640.818,
        490.483,
        318.578,
        9.682,
        10.035,
        9.134,
        8.554,
        9.274,
        9.463,
        9.501,
        10.036,
        9.211,
        8.265,
    ]
)


def test_hildebrand_sekhon():
    noise_params = hildebrand_sekhon.estimate_noise_hs74(SAMPLE_SPECTRUM)
    mean, thresh, var, nnoise = noise_params
    assert_almost_equal(mean, 10, 0)
    assert_almost_equal(thresh, 12, 0)
    assert_almost_equal(var, 0.79, 2)
    assert nnoise == 27



""" Unit Tests for Py-ART's graph/cm_colorblind.py module. """


import matplotlib

from pyart.graph import cm_colorblind


def test_colormaps_exist():
    assert isinstance(cm_colorblind.HomeyerRainbow, matplotlib.colors.Colormap)
    assert isinstance(cm_colorblind.HomeyerRainbow_r, matplotlib.colors.Colormap)
    assert isinstance(cm_colorblind.balance, matplotlib.colors.Colormap)
    assert isinstance(cm_colorblind.balance_r, matplotlib.colors.Colormap)
    assert isinstance(cm_colorblind.ChaseSpectral, matplotlib.colors.Colormap)
    assert isinstance(cm_colorblind.ChaseSpectral_r, matplotlib.colors.Colormap)
    assert isinstance(cm_colorblind.SpectralExtended, matplotlib.colors.Colormap)
    assert isinstance(cm_colorblind.SpectralExtended_r, matplotlib.colors.Colormap)


def test_colormaps_registered():
    cmap = matplotlib.colormaps.get_cmap("pyart_HomeyerRainbow")
    assert isinstance(cmap, matplotlib.colors.Colormap)

    cmap = matplotlib.colormaps.get_cmap("pyart_HomeyerRainbow_r")
    assert isinstance(cmap, matplotlib.colors.Colormap)

    cmap = matplotlib.colormaps.get_cmap("pyart_balance")
    assert isinstance(cmap, matplotlib.colors.Colormap)

    cmap = matplotlib.colormaps.get_cmap("pyart_balance_r")
    assert isinstance(cmap, matplotlib.colors.Colormap)

    cmap = matplotlib.colormaps.get_cmap("pyart_ChaseSpectral")
    assert isinstance(cmap, matplotlib.colors.Colormap)

    cmap = matplotlib.colormaps.get_cmap("pyart_ChaseSpectral_r")
    assert isinstance(cmap, matplotlib.colors.Colormap)

    cmap = matplotlib.colormaps.get_cmap("pyart_SpectralExtended")
    assert isinstance(cmap, matplotlib.colors.Colormap)

    cmap = matplotlib.colormaps.get_cmap("pyart_SpectralExtended_r")
    assert isinstance(cmap, matplotlib.colors.Colormap)



"""
Simple moment calculations.

"""

import numpy as np
from scipy import ndimage

from ..config import get_field_name, get_metadata
from ..core.transforms import antenna_to_cartesian
from ..util import angular_texture_2d


def calculate_snr_from_reflectivity(
    radar, refl_field=None, snr_field=None, toa=25000.0
):
    """
    Calculate the signal to noise ratio, in dB, from the reflectivity field.

    Parameters
    ----------
    radar : Radar
        Radar object from which to retrieve reflectivity field.
    refl_field : str, optional
        Name of field in radar which contains the reflectivity.
        None will use the default field name in the Py-ART configuration file.
    snr_field : str, optional
        Name to use for snr metadata. None will use the default field name
        in the Py-ART configuration file.
    toa : float, optional
        Height above which to take noise floor measurements, in meters.

    Returns
    -------
    snr : field dictionary
        Field dictionary containing the signal to noise ratio.

    """
    if refl_field is None:
        refl_field = get_field_name("reflectivity")
    if snr_field is None:
        snr_field = get_field_name("signal_to_noise_ratio")

    range_grid = (
        np.meshgrid(radar.range["data"], np.ma.ones(radar.time["data"].shape))[0] + 1.0
    )

    # remove range scale.. This is basically the radar constant scaled dBm
    pseudo_power = radar.fields[refl_field]["data"] - 20.0 * np.log10(
        range_grid / 1000.0
    )

    # Noise floor estimate
    # 25km.. should be no scatterers, not even planes, this high
    # we could get undone by AP though.. also sun
    rg, azg = np.meshgrid(radar.range["data"], radar.azimuth["data"])
    rg, eleg = np.meshgrid(radar.range["data"], radar.elevation["data"])
    _, _, z = antenna_to_cartesian(rg / 1000.0, azg, eleg)  # XXX: need to fix

    points_above = np.where(z > toa)
    noise_floor_estimate = pseudo_power[points_above].mean()

    snr_dict = get_metadata(snr_field)
    snr_dict["data"] = pseudo_power - noise_floor_estimate
    return snr_dict


def compute_noisedBZ(nrays, noisedBZ_val, _range, ref_dist, noise_field=None):
    """
    Computes noise in dBZ from reference noise value.

    Parameters
    ----------
    nrays : int
        Number of rays in the reflectivity field.
    noisedBZ_val : float
        Estimated noise value in dBZ at reference distance.
    _range : np array of floats
        Range vector in m.
    ref_dist : float
        Reference distance in Km.
    noise_field : str, optional
        Name of the noise field.

    Returns
    -------
    noisedBZ : dict
        The noise field.

    """
    # parse the field parameters
    if noise_field is None:
        noise_field = get_field_name("noisedBZ_hh")

    noisedBZ_vec = noisedBZ_val + 20.0 * np.ma.log10(1e-3 * _range / ref_dist)

    noisedBZ = get_metadata(noise_field)
    noisedBZ["data"] = np.tile(noisedBZ_vec, (nrays, 1))

    return noisedBZ


def compute_snr(radar, refl_field=None, noise_field=None, snr_field=None):
    """
    Computes SNR from a reflectivity field and the noise in dBZ.

    Parameters
    ----------
    radar : Radar
        Radar object
    refl_field : str, optional
        Name of the reflectivity field to use.
    noise_field : str, optional
        Name of the noise field to use.
    snr_field : str, optional
        Name of the SNR field.

    Returns
    -------
    snr : dict
        The SNR field.

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name("reflectivity")
    if noise_field is None:
        noise_field = get_field_name("noisedBZ_hh")
    if snr_field is None:
        snr_field = get_field_name("signal_to_noise_ratio")

    # extract fields from radar
    if refl_field in radar.fields:
        refl = radar.fields[refl_field]["data"]
    else:
        raise KeyError("Field not available: " + refl_field)
    if noise_field in radar.fields:
        noisedBZ = radar.fields[noise_field]["data"]
    else:
        raise KeyError("Field not available: " + noise_field)

    snr_data = refl - noisedBZ

    snr = get_metadata(snr_field)
    snr["data"] = snr_data

    return snr


def compute_l(radar, rhohv_field=None, l_field=None):
    """
    Computes Rhohv in logarithmic scale according to L=-log10(1-RhoHV).

    Parameters
    ----------
    radar : Radar
        Radar object.
    rhohv_field : str, optional
        Name of the RhoHV field to use.
    l_field : str, optional
        Name of the L field.

    Returns
    -------
    l_field_out : dict
        L field.

    """
    # parse the field parameters
    if rhohv_field is None:
        rhohv_field = get_field_name("cross_correlation_ratio")
    if l_field is None:
        l_field = get_field_name("logarithmic_cross_correlation_ratio")

    # extract rhohv field from radar
    if rhohv_field in radar.fields:
        rhohv = radar.fields[rhohv_field]["data"]
    else:
        raise KeyError("Field not available: " + rhohv_field)

    rhohv[rhohv >= 1.0] = 0.9999
    l_data = -np.ma.log10(1.0 - rhohv)

    l_field_out = get_metadata(l_field)
    l_field_out["data"] = l_data

    return l_field_out


def compute_cdr(radar, rhohv_field=None, zdr_field=None, cdr_field=None):
    """
    Computes the Circular Depolarization Ratio.

    Parameters
    ----------
    radar : Radar
        Radar object.
    rhohv_field : str, optional
        Name of the RhoHV field.
    zdr_field : str, optional
        Name of the ZDR field.
    cdr_field : str, optional
        Name of the CDR field.

    Returns
    -------
    cdr : dict
        CDR field.

    """
    # parse the field parameters
    if rhohv_field is None:
        rhohv_field = get_field_name("cross_correlation_ratio")
    if zdr_field is None:
        zdr_field = get_field_name("differential_reflectivity")
    if cdr_field is None:
        cdr_field = get_field_name("circular_depolarization_ratio")

    # extract fields from radar
    if rhohv_field in radar.fields:
        rhohv = radar.fields[rhohv_field]["data"]
    else:
        raise KeyError("Field not available: " + rhohv_field)
    if zdr_field in radar.fields:
        zdrdB = radar.fields[zdr_field]["data"]
    else:
        raise KeyError("Field not available: " + zdr_field)

    zdr = np.ma.power(10.0, 0.1 * zdrdB)

    cdr_data = 10.0 * np.ma.log10(
        (1.0 + 1.0 / zdr - 2.0 * rhohv * np.ma.sqrt(1.0 / zdr))
        / (1.0 + 1.0 / zdr + 2.0 * rhohv * np.ma.sqrt(1.0 / zdr))
    )

    cdr = get_metadata(cdr_field)
    cdr["data"] = cdr_data

    return cdr


def calculate_velocity_texture(
    radar, vel_field=None, wind_size=4, nyq=None, check_nyq_uniform=True
):
    """
    Derive the texture of the velocity field.

    Parameters
    ----------
    radar: Radar
        Radar object from which velocity texture field will be made.
    vel_field : str, optional
        Name of the velocity field. A value of None will force Py-ART to
        automatically determine the name of the velocity field.
    wind_size : int, optional
        The size of the window to calculate texture from. The window is
        defined to be a square of size wind_size by wind_size.
    nyq : float, optional
        The nyquist velocity of the radar. A value of None will force Py-ART
        to try and determine this automatically.
    check_nyquist_uniform : bool, optional
        True to check if the Nyquist velocities are uniform for all rays
        within a sweep, False will skip this check. This parameter is ignored
        when the nyq parameter is not None.

    Returns
    -------
    vel_dict: dict
        A dictionary containing the field entries for the radial velocity
        texture.

    """
    # Parse names of velocity field
    if vel_field is None:
        vel_field = get_field_name("velocity")

    # Allocate memory for texture field
    vel_texture = np.zeros(radar.fields[vel_field]["data"].shape)

    # If an array of nyquist velocities is derived, use different
    # nyquist velocites for each sweep in texture calculation according to
    # the nyquist velocity in each sweep.

    if nyq is None:
        # Find nyquist velocity if not specified
        nyq = [
            radar.get_nyquist_vel(i, check_nyq_uniform) for i in range(radar.nsweeps)
        ]
        for i in range(0, radar.nsweeps):
            start_ray, end_ray = radar.get_start_end(i)
            inds = range(start_ray, end_ray)
            vel_sweep = radar.fields[vel_field]["data"][inds]
            vel_texture[inds] = angular_texture_2d(vel_sweep, wind_size, nyq[i])
    else:
        vel_texture = angular_texture_2d(
            radar.fields[vel_field]["data"], wind_size, nyq
        )
    vel_texture_field = get_metadata("velocity")
    vel_texture_field["long_name"] = "Doppler velocity texture"
    vel_texture_field["standard_name"] = (
        "texture_of_radial_velocity" + "_of_scatters_away_from_instrument"
    )
    vel_texture_field["data"] = ndimage.median_filter(
        vel_texture, size=(wind_size, wind_size)
    )
    return vel_texture_field



""" Unit Tests for Py-ART's correct/unwrap.py module. """

# python test_dealias.py
# to recreate a dealias_plot.png file showing the before and after doppler
# velocities

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_almost_equal

import pyart

REF_DATA = [
    0.5,
    1.5,
    2.5,
    3.5,
    4.5,
    5.5,
    6.5,
    7.5,
    8.5,
    9.5,
    10.5,
    11.5,
    12.5,
    13.5,
    12.5,
    11.5,
    10.5,
    9.5,
    8.5,
    7.5,
    6.5,
    5.5,
    4.5,
    3.5,
    2.5,
    1.5,
    0.5,
]


def test_dealias_unwrap_phase_ray():
    radar, dealias_vel = perform_dealias("ray")
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False


def test_dealias_unwrap_phase_sweep():
    radar, dealias_vel = perform_dealias("sweep")
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False


def test_dealias_unwrap_phase_volume():
    radar, dealias_vel = perform_dealias("volume")
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False


def test_set_limits():

    radar, dealias_vel = perform_dealias("ray", set_limits=True)
    assert "valid_min" in dealias_vel
    assert_almost_equal(dealias_vel["valid_min"], -30.0)
    assert "valid_max" in dealias_vel
    assert_almost_equal(dealias_vel["valid_max"], 30.0)

    radar, dealias_vel = perform_dealias("ray", set_limits=False)
    assert "valid_min" not in dealias_vel
    assert "valid_max" not in dealias_vel


def test_dealias_unwrap_phase_no_gatefilter():
    radar, dealias_vel = perform_dealias("sweep", gatefilter=False)
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False


def test_dealias_unwrap_phase_explicit_gatefilter():
    radar = pyart.testing.make_velocity_aliased_radar()
    gf = pyart.correct.GateFilter(radar)
    radar.fields["velocity"]["data"][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_unwrap_phase(radar, gatefilter=gf)
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False


def test_dealias_unwrap_phase_masked_field_sweep():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields["velocity"]["data"]
    radar.fields["velocity"]["data"] = np.ma.masked_array(vdata)
    radar.fields["velocity"]["data"][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields["velocity"]["data"][180, 25] = np.ma.masked
    dealias_vel = pyart.correct.dealias_unwrap_phase(radar, unwrap_unit="volume")
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False
    assert np.ma.is_masked(dealias_vel["data"][180, 25])
    assert not np.ma.is_masked(dealias_vel["data"][180, 24])


def test_dealias_unwrap_phase_masked_field_volume():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields["velocity"]["data"]
    radar.fields["velocity"]["data"] = np.ma.masked_array(vdata)
    radar.fields["velocity"]["data"][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields["velocity"]["data"][180, 25] = np.ma.masked
    dealias_vel = pyart.correct.dealias_unwrap_phase(radar, unwrap_unit="volume")
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False
    assert np.ma.is_masked(dealias_vel["data"][180, 25])
    assert not np.ma.is_masked(dealias_vel["data"][180, 24])


def test_dealias_unwrap_phase_masked_field_ray():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields["velocity"]["data"]
    radar.fields["velocity"]["data"] = np.ma.masked_array(vdata)
    radar.fields["velocity"]["data"][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields["velocity"]["data"][180, 25] = np.ma.masked
    dealias_vel = pyart.correct.dealias_unwrap_phase(radar, unwrap_unit="ray")
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False
    assert np.ma.is_masked(dealias_vel["data"][180, 25])
    assert not np.ma.is_masked(dealias_vel["data"][180, 24])


def test_dealias_unwrap_phase_rhi_sweep():
    radar = pyart.testing.make_velocity_aliased_rhi_radar()
    radar.fields["velocity"]["data"][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_unwrap_phase(radar)
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False


def test_dealias_unwrap_phase_rhi_volume():
    radar = pyart.testing.make_velocity_aliased_rhi_radar()
    radar.fields["velocity"]["data"][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_unwrap_phase(radar, unwrap_unit="volume")
    assert_allclose(dealias_vel["data"][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel["data"][13]) is False


def test_dealias_keep_original():
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.fields["velocity"]["data"][180, 5] = 88
    gf = pyart.filters.GateFilter(radar)
    gf.exclude_above("velocity", 40)

    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, gatefilter=gf, keep_original=False
    )
    assert np.ma.is_masked(dealias_vel["data"][180, 5]) is True

    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, gatefilter=gf, keep_original=True
    )
    assert_almost_equal(dealias_vel["data"][180, 5], 88)
    assert np.ma.is_masked(dealias_vel["data"][180, 5]) is False


def test_is_radar_sweep_aligned():
    radar = pyart.testing.make_empty_ppi_radar(3, 4, 2)
    radar.scan_type = "rhi"
    radar.elevation["data"][:] = [1, 2, 3, 4, 1, 2, 3, 4]
    assert pyart.correct.unwrap._is_radar_sweep_aligned(radar)

    radar.elevation["data"][:] = [1, 2, 3, 4, 2, 3, 4, 5]
    assert not pyart.correct.unwrap._is_radar_sweep_aligned(radar)

    # raises ValueError
    radar = pyart.testing.make_empty_ppi_radar(3, 4, 2)
    radar.scan_type = "fuzz"
    pytest.raises(ValueError, pyart.correct.unwrap._is_radar_sweep_aligned, radar)


def test_is_sweep_sequential():
    radar = pyart.testing.make_empty_ppi_radar(3, 4, 2)
    radar.scan_type = "vpt"
    assert pyart.correct.unwrap._is_sweep_sequential(radar, 0)


def test_dealias_unwrap_phase_raises():

    # invalid unwrap_unit
    radar = pyart.testing.make_velocity_aliased_radar()
    pytest.raises(
        ValueError, pyart.correct.dealias_unwrap_phase, radar, unwrap_unit="fuzz"
    )

    # no explicit nyquist
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.instrument_parameters = None
    pytest.raises(LookupError, pyart.correct.dealias_unwrap_phase, radar)

    # non-sequential
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.azimuth["data"][10] = 190.0
    pytest.raises(ValueError, pyart.correct.dealias_unwrap_phase, radar)

    # non-aligned sweeps
    radar = pyart.testing.make_empty_ppi_radar(1, 10, 2)
    radar.fields["velocity"] = {"data": np.zeros((20, 1))}
    pytest.raises(
        ValueError,
        pyart.correct.dealias_unwrap_phase,
        radar,
        nyquist_vel=10,
        unwrap_unit="volume",
    )

    # non-cubic radar
    radar = pyart.testing.make_empty_ppi_radar(1, 10, 2)
    radar.fields["velocity"] = {"data": np.zeros((20, 1))}
    radar.azimuth["data"][-10:] = range(10)
    radar.sweep_end_ray_index["data"][-1] = 18
    pytest.raises(
        ValueError,
        pyart.correct.dealias_unwrap_phase,
        radar,
        nyquist_vel=10,
        unwrap_unit="volume",
    )

    # invalid scan type
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.scan_type = "fuzz"
    pytest.raises(ValueError, pyart.correct.dealias_unwrap_phase, radar)

    # invalid scan type
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.scan_type = "fuzz"
    pytest.raises(ValueError, pyart.correct.dealias_unwrap_phase, radar)


def perform_dealias(unwrap_unit="sweep", **kwargs):
    """Perform velocity dealiasing on reference data."""
    radar = pyart.testing.make_velocity_aliased_radar()
    # speckling that will not be not be dealiased.
    radar.fields["velocity"]["data"][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, unwrap_unit=unwrap_unit, **kwargs
    )
    return radar, dealias_vel


def test_fillvalue(**kwargs):
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.fields["velocity"]["_FillValue"] = 8888.0

    # ensure that at least one gate is excluded and should be masked
    radar.fields["velocity"]["data"][100, 25] = 44
    gf = pyart.filters.GateFilter(radar)
    gf.exclude_above("velocity", 40)

    dealias_vel = pyart.correct.dealias_unwrap_phase(radar, gatefilter=gf)
    assert isinstance(dealias_vel["data"], np.ma.MaskedArray)
    assert np.ma.is_masked(dealias_vel["data"][100, 25])

    assert "_FillValue" in dealias_vel
    assert dealias_vel["_FillValue"] == 8888.0
    assert dealias_vel["data"].fill_value == 8888.0


if __name__ == "__main__":

    radar, dealias_vel = perform_dealias()
    radar.fields["dealiased_velocity"] = dealias_vel

    # print out results
    print("ray 13 velocitites before dealias:")
    print(radar.fields["velocity"]["data"][13])
    print("ray 13 velocities after dealias:")
    print(radar.fields["dealiased_velocity"]["data"][13])

    # create plot
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=[5, 10])
    ax1 = fig.add_subplot(2, 1, 1)
    ax2 = fig.add_subplot(2, 1, 2)
    rd = pyart.graph.RadarDisplay(radar)
    rd.plot_ppi("velocity", 0, ax=ax1, colorbar_flag=False, title="", vmin=-10, vmax=10)
    rd.plot_ppi(
        "dealiased_velocity",
        0,
        ax=ax2,
        colorbar_flag=False,
        title="",
        vmin=-10,
        vmax=20,
    )
    fig.savefig("dealias_plot.png")
